###########################################################################
# Monte M. Goode, LBNL
# See Copyright for copyright notice!
###########################################################################

# $Id: Decorator.py,v 1.28 2004/03/22 01:32:50 mgoode Exp $

# replacement type writer for globus specific stuff

import re, sys
from ZSI.wsdl2python import SchemaDescription

ID1 = '    '
ID2 = 2*ID1
ID3 = 3*ID1
ID4 = 4*ID1
ID5 = 5*ID1
ID6 = 6*ID1

DEBUG = False

def print_debug(msg, arg=None):
    if DEBUG and (arg != None):
        print '<DEC> %s:' % msg, arg
    elif DEBUG:
        print '<DEC> %s' % msg


# these lists hold additional imports for the generated code

typeimports = [
    '\nfrom ZSI import _seqtypes',
    '\nimport pyGridWare',
    '\nfrom pyGridWare.ogsi import TCmodified',
    '\nfrom pyGridWare.ogsi.TCmodified import AddNamespaceContainer',
    '\nfrom pyGridWare.ogsi.TCmodified import AnyElement, Struct',
    '\nZSI.TCcompound.Struct = TCmodified.Struct'
    ]

clientimports = [
    '\nimport pyGridWare',
    '\nfrom pyGridWare.ogsi.security.SecurityClient import SecurityBinding',
    '\nclass client:',
    '\n%sBinding = SecurityBinding' % ID1,
    ]

# replacement type interpreter to handle some non-zsi stuff

from ZSI.typeinterpreter import BaseTypeInterpreter
from TCmodified import AnyElement
from TCmodified import QName

class ExtendedTypeInterpreter(BaseTypeInterpreter):
    def __init__(self):
        BaseTypeInterpreter.__init__(self)
        print_debug('ExtendedTypeInterpreter initialized')
        self._type_list.append(AnyElement)
        self._type_list.append(QName)

# subclasses the TypeWriter and adds functionality where needed.

class OGSIDecorator( SchemaDescription.TypeWriter ):
    def __init__(self):
        print_debug( 'OGSI Decorator Initialized' )
        SchemaDescription.TypeWriter.__init__(self)

        self.processingGlobalDef = True

        # substitute a new TypeInterpreter (both stylish and sketchy!)
        sys.modules['ZSI.wsdlInterface'].__dict__['AdapterBase'].__dict__['bti'] = ExtendedTypeInterpreter()


    def fromType(self, myType, parentRef):
        print_debug( 'fromType()' )
        # these modifications require a bit of explaination.  the single
        # parent class (passed in as the weakref 'parentRef') creates
        # multiple typeWriters (this class' baseclass) in a loop.  here
        # modifications are being made to this single parent class.
        # so the change is being tested for to see if it has been made
        # already.

        # these changes add class vars to the top of the namespace class.
        if not re.search( 'kw_global', parentRef().header):
            parent = parentRef()
            prefix = []
            for s in parent.nsh.getModuleName( myType.getTargetNamespace()
                                               ).split('_'):
                if len(s) > 4:
                    prefix.append(s[0:4])
                else:
                    prefix.append(s)
                    
            parent.header += '\n%skw_global = {' % ID1
            parent.header += "\n%s'inorder': 0," % ID2
            parent.header += "\n%s'unique': 1," % ID2
            parent.header += "\n%s'typed': 1," % ID2
            parent.header += "\n%stargetNamespace: '%s'" \
                             % ( ID2, '_'.join(prefix) )
            parent.header += '\n%s}' % ID2
            parent.header += '\n%skw_local = kw_global.copy()' % ID1
            parent.header += "\n%skw_local['typed'] = 0" % ID1
            parent.header += "\n\n%slocalParameters = ('repeatable', 'optional', 'type')" % ID1

        # this change adds code after the namespace class is generated.
        if not parentRef().last:
            parent = parentRef()
            parent.last += '\n# additional wildcard code'
            parent.last += '\nTCmodified.AddNamespaceContainer(namespace="%s", container=%s)' % (myType.getTargetNamespace(), parent.nsh.getModuleName(myType.getTargetNamespace()))

        # now the true baseclass is called and we are back to our
        # regularly scheduled TypeWriter....
        SchemaDescription.TypeWriter.fromType(self, myType, parentRef)

    def _fromSimpleType(self, tp):
        print_debug( '_fromSimpleType()' )

        alias = self.nsh.getAlias( tp.getTargetNamespace())
        
        SchemaDescription.TypeWriter._fromSimpleType(self, tp)

        self.initdef.set('\n%sdef __init__(self, pname, **kw):' % ID2)

        self.initcode.set('\n%sif type(pname) in _seqtypes:' % ID3)
        self.initcode.write('\n%saname = "_%%s" %% pname[1]' % ID4)
        self.initcode.write('\n%selse:' % ID3)
        self.initcode.write('\n%saname = "_%%s" %% pname' % ID4)
        self.initcode.write('\n%spname = (self.schema,pname)' % ID4)

        self.initcode.write('\n\n%skw.update(%s.kw_global)' % (ID3, alias))

        self.initcode.write("\n\n%sxmlns = ''" % ID3 )
        self.initcode.write("\n%sif not kw.has_key(self.schema):" % ID3 )
        self.initcode.write("\n%sxmlns = ' xmlns:%%s=\"%%s\"' %% (%s.kw_global[self.schema],self.schema)" % (ID4, alias))

        
        self.initcode.write('\n%soname = "%%s:%%s%%s" %% (kw[pname[0]],pname[1], xmlns)' % ID3)
        
        self.initcode.write("\n\n%sif not kw.has_key('type'):" % ID3)
        self.initcode.write("\n%skw['type'] = self.type" % ID4)

        # tweek the base ctor
        base = re.split( ',', self.basector.getvalue() )[0].strip()
        self.basector.set('\n\n%s%s,' % (ID3, base))
        self.basector.write('\n%spname=pname,' % ID4)
        self.basector.write('\n%saname=aname,' % ID4)
        self.basector.write('\n%soname=oname,' % ID4)
        self.basector.write('\n%soptional = 1,' % ID4)
        self.basector.write('\n%srepeatable = 1,' % ID4)
        self.basector.write('\n%s**kw)' % ID4)
        

        self.postpend.set("\n%sself.tag = '%%s:%%s' %%(kw[kw['type'][0]],kw['type'][1])" % ID3)
        self.postpend.write('\n\n%s%s_Def.schema = targetNamespace' \
                            % (ID1, tp.getName()))
        self.postpend.write("\n%s%s_Def.type = (targetNamespace, '%s')" \
                            % ( ID1, tp.getName(), tp.getName()))

    def _fromWildCard(self, tp):
        print_debug( '_fromWildCard()' )
        SchemaDescription.TypeWriter._fromWildCard(self, tp)

    def _fromAttribute(self, tp):
        print_debug( '_fromAttribute()' )
        SchemaDescription.TypeWriter._fromAttribute(self, tp)

    def _fromElement(self, tp):
        print_debug( '_fromElement()' )
        SchemaDescription.TypeWriter._fromElement(self, tp)

    def _elementSimpleType(self, tp, etp):
        print_debug( '_elementSimpleType()' )

        alias = self.nsh.getAlias( tp.getTargetNamespace())
        SchemaDescription.TypeWriter._elementSimpleType(self, tp, etp)

        self.classvar.set("\n%sliteral = '%s'\n" % (ID2, tp.getName()))

        self.initdef.set('\n%sdef __init__(self, name=None, ns=None, **kw):'\
                         % ID2 )

        self.initcode.write("\n\n%sxmlns = ''" % ID3 )
        self.initcode.write("\n%sif not kw.has_key(self.schema):" % ID3 )
        self.initcode.write("\n%sxmlns = ' xmlns:%%s=\"%%s\"' %% (%s.kw_global[self.schema],self.schema)" % (ID4, alias))


        self.initcode.write('\n\n%skw.update(%s.kw_global)' % (ID3, alias))
        self.initcode.write('\n%spname=(ns,name)' % ID3)
        self.initcode.write("\n%saname='_%%s' %% name" % ID3)
        self.initcode.write("\n%soname = '%%s:%%s%%s' %% (kw[ns],name,xmlns)"\
                            % ID3)

        base = re.split( ',', self.basector.getvalue() )[0].strip()

        self.basector.set('\n\n%s%s,' % (ID3, base))
        self.basector.write('\n%spname=(ns,name),' % ID4)
        self.basector.write('\n%saname=aname,' % ID4)
        self.basector.write('\n%soname=oname,' % ID4)
        self.basector.write('\n%s**kw)' % ID4)

        self.postpend.set('\n\n%s%s_Dec.schema = targetNamespace' \
                          %(ID1, tp.getName()))


    def _elementComplexType(self, tp, etp):
        print_debug( '_elementComplexType()' )

        localDef = False

        if not etp.getName():
            self.processingGlobalDef = False

        SchemaDescription.TypeWriter._elementComplexType(self, tp, etp)

        if not self.processingGlobalDef:
            # this is all now handled mostly by the complex type
            # handling code.  no double classes anymore.

            self.classvar.set('\n%sliteral = "%s"\n' % (ID2, etp.getName()))
            
            ctorcode = self.initcode.getvalue()

            self.initdef.set('\n%sdef __init__(self, name=None, ns=None, **kw):' % ID2)

            self.initcode.set('\n%sname = self.__class__.literal'% ID3)
            self.initcode.write('\n%sns = ns or self.__class__.schema' % ID3)
            self.initcode.write('\n%spname = (ns,name)\n' % ID3)
            self.initcode.write( ctorcode )

            self.postpend.set('\n\n%s%s_Dec.schema = targetNamespace' \
                              % ( ID1, etp.getName()))

        else:
            # remove schema
            self.classvar.set('\n%sliteral = "%s"\n' % ( ID2, tp.getName()))
            # replace initdef
            self.initdef.set('\n%sdef __init__(self, name=None, ns=None, **kw):' % ID2 )
            # add logic
            self.initcode.write('\n\n%skw.update(%s.kw_global)'\
                                % (ID3,
                                   self.nsh.getAlias(tp.getTargetNamespace())))
            self.initcode.write('\n%spname = (ns,name)' % ID3)
            # doctor up the base ctors and etc
            nsp = etp.getTargetNamespace()
            self.basector.set('\n\n%s%s.%s.__init__(self, pname=pname, **kw)'\
                              %(ID3,
                                self.nsh.getAlias(nsp),
                                etp.getName() + '_Def'))
            self.postpend.set('\n%sself.typecode = %s.%s(pname=pname, **kw)' \
                              % (ID3, self.nsh.getAlias(nsp),
                                 etp.getName() + '_Def'))
            self.postpend.write('\n\n%s%s_Dec.schema = targetNamespace' \
                                % (ID1, tp.getName()))
            
    def _fromComplexType(self, tp):
        print_debug( '_fromComplexType()' )
        SchemaDescription.TypeWriter._fromComplexType(self, tp)

        self.classvar.set('')

        self.classvar.write("\n%sschema = '%s'" \
                            % (ID2, tp.getTargetNamespace()))
        self.classvar.write("\n%stype = (schema, '%s')\n" \
                            % (ID2, tp.getName()))

        
        print_debug( 'exiting _fromComplexType()')

    def _complexTypeComplexContent(self, tp):
        print_debug( '_complexTypeComplexContent()' )
        SchemaDescription.TypeWriter._complexTypeComplexContent(self, tp)

        dt = tp.getDerivedTypes()
        tc = dt.getTypeclass()

        alias = self.nsh.getAlias( tp.getTargetNamespace())

        if not tc:
            if dt.getDerivation():
                # we have a derivation
                self.initdef.set('\n%sdef __init__(self, pname, **kw):' % ID2)
                # snatch the TCList
                tclist = re.search( '(TCList\s=.*)$', self.initcode.getvalue(),
                                    re.MULTILINE ).group(1)
                self.initcode.set('\n%sif pname not in _seqtypes:' % ID3)
                self.initcode.write('\n%spname = (self.schema, pname)' % ID4)

                self.initcode.write('\n\n%skw.update(%s.kw_global)' % (ID3, alias))
                self.initcode.write('\n%skwp = kw.copy()' % ID3)
                self.initcode.write('\n%sfor k in %s.localParameters: kwp.pop(k,None)' % (ID3, alias))
                self.initcode.write("\n%sif not kw.has_key('type'):" % ID3)
                self.initcode.write("\n%skw['type'] = self.type" % ID4)

                self.initcode.write('\n\n%s%s' % (ID3, tclist))

                base = re.split( ',', self.basector.getvalue() )[0].strip()

                self.basector.set('\n%s%s, pname=pname, **kw)' % (ID3, base))

                if dt.isExtension():
                    print_debug('extending')
                    self.postpend.set('\n%s# extending...' % ID3)
                    self.postpend.write('\n%sself.setDerivedTypeContents(extensions=TCList)' % ID3 )
                elif dt.isRestriction():
                    print_debug('restricting')
                    self.postpend.set('\n#s resticting...' % ID3)
                    self.postpend.write('\n%sself.setDerivedTypeContents(restrictions=TCList)' % ID3 )
                
        elif '%s' % tc == 'ZSI.TCcompound.Array':
            # a soapenc array
            pass
        else:
            # nothing, an exception will have been thrown in the superclass
            pass

    def _complexTypeSimpleContent(self, tp):
        print_debug( '_complexTypeSimpleContent()' )
        SchemaDescription.TypeWriter._complexTypeSimpleContent(self, tp)

        alias = self.nsh.getAlias( tp.getTargetNamespace())

        self.initdef.set('\n%sdef __init__(self, pname, **kw):' % ID2)

        self.initcode.set('\n%sif type(pname) in _seqtypes:' % ID3)
        self.initcode.write('\n%saname = "_%%s" %% pname[1]' % ID4)
        self.initcode.write('\n%selse:' % ID3)
        self.initcode.write('\n%saname = "_%%s" %% pname' % ID4)
        self.initcode.write('\n%spname = (self.schema,pname)' % ID4)

        self.initcode.write("\n\n%sxmlns = ''" % ID3 )
        self.initcode.write("\n%sif not kw.has_key(self.schema):" % ID3 )
        self.initcode.write("\n%sxmlns = ' xmlns:%%s=\"%%s\"' %% (%s.kw_global[self.schema],self.schema)" % (ID4, alias))


        self.initcode.write('\n\n%skw.update(%s.kw_global)' % (ID3, alias))
        self.initcode.write('\n%soname = "%%s:%%s%%s" %% (kw[pname[0]],pname[1],xmlns)' % ID3)
        
        self.initcode.write("\n\n%sif not kw.has_key('type'):" % ID3)
        self.initcode.write("\n%skw['type'] = self.type" % ID4)

        # tweek the base ctor
        base = re.split( ',', self.basector.getvalue() )[0].strip()
        self.basector.set('\n\n%s%s,' % (ID3, base))
        self.basector.write('\n%spname=pname,' % ID4)
        self.basector.write('\n%saname=aname,' % ID4)
        self.basector.write('\n%soname=oname,' % ID4)
        self.basector.write('\n%s**kw)' % ID4)
        

        self.postpend.set("\n%sself.tag = '%%s:%%s' %%(kw[kw['type'][0]],kw['type'][1])" % ID3)

    def _complexTypeAllOrSequence(self, tp, mg):
        print_debug( '_complexTypeAllOrSequence()' )
        r = SchemaDescription.TypeWriter._complexTypeAllOrSequence(self,tp,mg)
        return self._processTypeCodeStrings(r)

    def _complexTypeChoice(self, tp, mg):
        print_debug( '_complexTypeChoice()' )
        r = SchemaDescription.TypeWriter._complexTypeChoice(self, tp, mg)
        return self._processTypeCodeStrings(r)

    def _complexTypecodeLogic(self, typecodelist, tns ):
        print_debug( '_complexTypecodeLogic()' )

        alias = self.nsh.getAlias( tns )
        
        self.initdef.set('\n%sdef __init__(self, pname, **kw):' % ID2 )

        # grab the member inits
        members = self.initcode.getvalue()

        self.initcode.set('\n%sif type(pname) in _seqtypes:' % ID3)
        self.initcode.write('\n%saname = "_%%s" %% pname[1]' % ID4)
        self.initcode.write('\n%selse:' % ID3)
        self.initcode.write('\n%saname = "_%%s" %% pname' % ID4)
        self.initcode.write('\n%spname = (self.schema,pname)' % ID4)

        if self.processingGlobalDef:
            pass
            #self.initcode.write('\n\n%soname = "%%s:%%s" %% (kw[pname[0]],pname[1])' % ID3)
        
        self.initcode.write('\n\n%skw.update(%s.kw_global)' % (ID3, alias))
        self.initcode.write('\n%skwp = kw.copy()' % ID3)
        self.initcode.write('\n%sfor k in %s.localParameters: kwp.pop(k,None)' % (ID3, alias))
        if self.processingGlobalDef:
            self.initcode.write("\n%sif not kw.has_key('type'):" % ID3)
            self.initcode.write("\n%skw['type'] = self.type" % ID4)
            
        self.initcode.write( '\n%s' % members )

        self.basector.set('\n\n%sStruct.__init__(self,' % ID3)
        self.basector.write('\n%spyclass=self.__class__,' % ID4)
        self.basector.write('\n%sofwhat = %s,' % (ID4, typecodelist))
        self.basector.write('\n%spname=pname,' % ID4)
        self.basector.write('\n%saname=aname,' % ID4)

        extraFlags = ''

        if self.hasRepeatable:
            extraFlags += 'hasextras=1, '
        
        self.basector.write('\n%s%s**kw)' % (ID4, extraFlags))

        # none of the superclass method is needed here
        #SchemaDescription.TypeWriter._complexTypecodeLogic(self, typecodelist)

    def _complexTypeHandleAttributes(self, tp):
        print_debug( '_complexTypeHandleAttributes()' )
        r = SchemaDescription.TypeWriter._complexTypeHandleAttributes(self, tp)
        return r

    def _calculateOccurance(self, e):
        print_debug( '_calculateOccurance()' )
        r = SchemaDescription.TypeWriter._calculateOccurance(self, e)
        return r

    def typeDoc(self, occurs, objName, typeName):
        SchemaDescription.TypeWriter.typeDoc(self, occurs, objName, typeName)

    def _getWildcardTypecode(self, e, occurs):
        # this is a pure override - no logic from baseclass used
        return 'AnyElement(**kwp), '

    # the following methods are not in the base class interface

    def _processTypeCodeStrings(self, tcs):
        print_debug('_processTypeCodeStrings()')

        tcInList  = []
        tcOutList = []

        for t in tcs.split('),'):
            if t.strip():
                tmp = t.strip() + ')'
                tcInList.append( tmp )

        for t in tcInList:
            if re.match( 'ZSI\.', t ) or re.match('pyGridWare\.', t):
                print_debug('matched ZSI typecode: %s' % t)

                sObj = re.search("^([^\(]*)\(.*pname=\"([^\"]*)\".*aname=\"[^\"]*\",*([^\)]*)\)", t)

                if sObj:
                    argType = sObj.group(1)
                    argName = sObj.group(2)
                    flags   = sObj.group(3).strip()

                    tcOut = '%s(pname=(self.schema, "%s"), aname="_%s", oname="%%s:%s" %%kw[self.schema], ' % ( argType, argName, argName, argName)
                    
                    if flags.strip():
                        tcOut += '%s, **kwp)' % flags
                    else:
                        tcOut += '**kwp)'
                        
                    print_debug(tcOut)
                    tcOutList.append(tcOut)
                else:
                    raise RuntimeError, 'failed to match zsi typecode: %s' % t

            elif re.match( 'ns\d+\.', t ) or \
                     re.match( 'self\.__class__', t):
                print_debug('matched schema-defined tc: %s' % t)

                # first catch element references for minimal processing
                if re.search("\(ns=ns", t):
                    print_debug('element reference encountered: %s' % t)
                    sObj = re.search("^([^\(]*)\(ns=ns([^\)]*)\)", t)
                    
                    if sObj:
                        klass = sObj.group(1)
                        flags = sObj.group(2).strip()
                    else:
                        raise RuntimeError, 'failed to match element ref %s'\
                              % t
                    tcOut = '%s(' % klass
                    
                    if flags.strip():
                        tcOut += '%s, **kwp)' % flags[2:]
                    else:
                        tcOut += '**kwp)'
                        
                    print_debug(tcOut)
                    
                    tcOutList.append(tcOut)
                    continue

                # if it's not an element ref, then continue....

                sObj = re.search("^([^\(]*)\(name=\"([^\"]*)\".*ns=ns,*([^\)]*)\)", t )

                if sObj:
                    argType = sObj.group(1)
                    argName = sObj.group(2)
                    flags   = sObj.group(3).strip()

                    tcOut = '%s(pname=(self.schema, "%s"), ' \
                            % (argType, argName)
                    
                    if flags.strip():
                        tcOut += '%s, **kwp)' % flags
                    else:
                        tcOut += '**kwp)'

                    print_debug( tcOut )
                    tcOutList.append(tcOut)
                else:
                    raise RuntimeError, 'failed to match defined typecode %s' \
                          % t
            elif re.match('AnyElement', t):
                # special case handled in this class - does not need massaging
                tcOutList.append(t)
                
            else:
                print_debug('typecode ID failed: %s' % t)
                raise RuntimeError, 'typecode processing failed: %s' % t

        return ', '.join(tcOutList) + ', '

        
        
