#
# Struct w/ wildcard element support
#

from ZSI import _copyright, _attrs, _children, _child_elements, \
        _inttypes, _stringtypes, _seqtypes, _find_arraytype, _find_href, \
        _find_type, \
        EvaluateException, SoapWriter
from ZSI.TC import TypeCode, String, Any, Void
try:
    from xml.ns import SCHEMA, SOAP, XMLNS
except:
    from ZSI.compat import SCHEMA, SOAP, XMLNS
import re, types, copy

from ZSI.TCcompound import Struct as ZSIStruct
from ZSI.TC import XML as ZSIXML
from ZSI.TCcompound import  _check_typecode_list, _position_pat, _offset_pat,\
    _find_arrayposition, _find_arrayoffset

from xml.dom.ext import SplitQName
import cStringIO as StringIO

from pyGridWare import getLogger
logger = getLogger(__name__)

DEBUG = 0
def print_debug(msg, arg=None):
    if DEBUG and (arg != None):
        print "<TCMOD> %s" %msg, arg
    elif DEBUG:
        print "<TCMOD> %s" %msg


def AddNamespaceContainer(namespace, container):
    """convience function for adding a types module for wildcard 
       and derived-type resolution.

       namespace -- 
       container -- container that contains type defintions and 
          element declarations.  Must be a module, class, or instance.
    """
    AnyElement.tns_module_class_dict[namespace] = container


def HasNamespaceContainer(namespace):
    return AnyElement.tns_module_class_dict.has_key(namespace)


def GetNamespaceContainer(namespace):
    return AnyElement.tns_module_class_dict[namespace]


class AnyElement(TypeCode):
    """XML Schema Element WildCard.
       class variables:
            -- tns_module_class_dict: This is only needed for searching 
                  class type containers.
    """
    tns_module_class_dict = {}
    tag = 'any'
    def __init__(self, namespaces=['#all'],pname=None, strip=1, textprotect=1, **kw):
        pname = ''
        self.optional = 1
        self.kw = kw
        TypeCode.__init__(self, pname=pname, **kw)
        self.namespaces = namespaces

    def getKW(self):
        self.kw.update({'optional':self.optional,
                        'typed':self.typed, 
                        'repeatable':self.repeatable, 
                        'unique':self.unique,
                        'nspname':self.nspname
                       })
        return self.kw


class AnyURI(String):
    parselist = [ (None,'string'), (None,'anyURI') ]
    seriallist = [ types.StringType, types.UnicodeType ]
    tag = 'anyURI'
    def __init__(self, pname=None, strip=1, textprotect=1, **kw):
        TypeCode.__init__(self, pname, **kw)
        if kw.has_key('resolver'): self.resolver = kw['resolver']
        self.strip = strip
        self.textprotect = textprotect

class QName(String):
    '''A QName type
    '''
    parselist = [ (None,'QName') ]
    tag = 'QName'

    def __init__(self, pname=None, strip=1, textprotect=1, **kw):
        TypeCode.__init__(self, pname, **kw)
        if kw.has_key('resolver'): self.resolver = kw['resolver']
        self.strip = strip
        self.textprotect = textprotect

    def serialize(self, sw, pyobj, name=None, attrtext='', **kw):
        """support QName serialization of a tuple (namespace, name), 
           use namespace as keyword {namespace:prefix} for prefix,
           else declare a prefix ns1.
        """
        if type(pyobj) in _seqtypes:
            ns,ncname = pyobj
            if kw.has_key(ns):
                pyobj = '%s:%s' %(kw[ns],ncname)
            else:
                name = '%s xmlns:ns1="%s"' %(name or pname, ns)
                pyobj = 'ns1:%s' %ncname
        String.serialize(self, sw, pyobj, name, attrtext, **kw)
        return


class Base64Binary(String):
    '''A Base64 encoded string.
    Base64Binary should go into TC.py, but we need to work it into the mix.
    class Base64Binary(ZSI.TC.Base64String): This declaration didn't work with OGSI
    '''
    parselist = [ (None,'base64Binary')]
    tag = 'base64Binary'


class Union(Any):
    """ serialize is easy, just rely on TC.Any.
        parse is hard, cant use type tag to figure out what 
        typecode to use.. essentially need to look at data in
        order to figure out how to parse it.  But many values
        could be both, so there is no way to truely handle this
        correctly.  One hack would be just to sequentially
        attempt to parse a value, and succeed if no exception
        is thrown.  String should be the last one tried, since
        any value could be a string.  Allow user to specify
        a function that will make this choice?
    """
    pass


class XML(ZSIXML):
    def parse(self, elt, ps):
        if not self.wrapped:
            if self.copyit: 
                return elt.cloneNode(1)
            return elt
        c = _child_elements(elt)
        if not c:
            href = _find_href(elt)
            if not href:
                if self.optional: return None
                raise EvaluateException('Embedded XML document missing',
                        ps.Backtrace(elt))
            if href[0] != '#':
                return ps.ResolveHREF(href, self)
            elt = ps.FindLocalHREF(href, elt)
            c = _child_elements(elt)
        if _find_encstyle(elt) != "":
            #raise EvaluateException('Embedded XML has unknown encodingStyle',
            #       ps.Backtrace(elt)
            pass
        if len(c) != 1:
            raise EvaluateException('Embedded XML has more than one child',
                    ps.Backtrace(elt))
        
        if self.copyit: return c[0].cloneNode(1)
        return c[0]


class Struct(ZSIStruct):
    '''A structure.
    '''
    def __init__(self, pyclass, ofwhat, pname=None, inorder=0, inline=0,
    mutable=1, hasextras=0, **kw):
        '''pyclass -- the Python class to hold the fields
        ofwhat -- a list of fields to be in the struct
        hasextras -- ignore extra input fields
        inorder -- fields must be in exact order or not
        inline -- don't href/id when serializing
        mutable -- object could change between multiple serializations
        type -- the (URI,localname) of the datatype
        '''
        ZSIStruct.__init__(self, pyclass, ofwhat, 
            pname, inorder, inline, mutable, hasextras, **kw)
        self.lenofwhat = len(ofwhat)
        if kw.has_key('typed'):
            self.typed = kw['typed']
        self.declared = kw.get('declared', [])
        self.serialization_nsdict = kw
        for key in ('aname','type','inline','mutable','pname','nspname', 
            'oname','typed','repeatable','optional','default','unique','declared'):
            self.serialization_nsdict.pop(key,None)
        self.attributes = {}

    def checktype(self, elt, ps):
        """If not exact type, check derived types.
        """
        checktype = ZSIStruct.checktype(self, elt, ps)
        if checktype not in [ self.type, (None,None) ]:
            container = GetNamespaceContainer(checktype[0])
            for c,v in container.__dict__.items():
                if hasattr(v, 'type') and v.type == checktype:
                    if not issubclass(v, self.__class__):
                        raise EvaluateException('Struct has wrong type', ps.Backtrace(elt))
                    self.hasextras = 1

        return checktype

    def findAttributeType(self, attribute, namespace=None, what=None):
        """ This method looks for the attribute type in the class variable
            attributeTypes, if not found it will look for attributeTypes by
            recursing with the base classes.  
            This is a simple method of allowing derived type extensions, 
            but currently this enables restrictions to act as extensions.

            Check attributeNS if namespace is provided.
            XXX This needs to changes namespaces should be indepedent, 
            currently can't specify two attributes of the same name in
            different namespaces.

            Solutions: 
                new class variable derivedType = 'extension | 'restriction'

            namespace -- check against what it is in attributeNS dict.

            what -- class instance or class definition w/ attributeType
                class variable.
            attribute -- what we are looking for in attributeType dict.

        """
        if not hasattr(self, 'attributeTypes') or not self.attributeTypes.has_key(attribute):
            return None

        if namespace:
            if self.attributeNS[attribute] != namespace:
                raise EvaluateError, 'bad attribute namespace'
        if what:
            if hasattr(what, 'attributeTypes') and \
                what.attributeTypes.has_key(attribute):
                return what.attributeTypes[attribute](attribute)
            if type(what) == types.ClassType:
                bases = what.__bases__
            else: 
                bases = what.__class__.__bases__

            for base in bases:
                attr_type = self.findAttributeType(attribute=attribute, namespace=namespace, what=base)
                if attr_type:
                    return attr_type
        else:
            if hasattr(self, 'attributeTypes') and \
                self.attributeTypes.has_key(attribute):
                return self.attributeTypes[attribute](attribute)
            for base in self.__class__.__bases__:
                attr_type = self.findAttributeType(attribute=attribute, namespace=namespace, what=base)
                if attr_type:
                    return attr_type
        return None

    def setSerializationAttribute(self, key, attribute):
        if not self.attributeTypes.has_key(key):
            raise EvaluateException, 'Attribute(%s) is not defined' %key
        self.attributes[key] = attribute

    def setSerializationElement(self, variable=None, typecode=None, nil=False):
        """Require all variables to start with single underscore.
           parameters:
            variable -- a type's element name, if type is an element then
               do not specify it. eg. _name

            typecode -- class definition used for variable, if not specified
               default to the matching what in self.ofwhat

            nil -- if set to True, then set variable to TC.Void.  Must be
               what in self.ofwhat must be nillable/optional.

        """
        nillable = True
        if variable and not variable.startswith('__'):
            if not variable.startswith('_'):
                aname = '_%s' %variable
            else:
                aname = variable

            if not typecode:
                for what in self.ofwhat:
                    if what.aname == aname:
                        break
                else:
                    raise EvaluateException, 'No typecode for aname %s in ofwhat' %aname
                pyclass = what.__class__
                nillable = what.optional
            elif type(typecode) == types.ClassType:
                pyclass = typecode
            else:
                raise EvaluateException, 'typecode must be a class definition'

            logger.info('Struct.setSerializationElement aname=%s, nspname=%s',aname,self.nspname)
            if nil:
                if not nillable:
                    logger.error('element %s is not nillable', pyclass.__name__)

                setattr(self, aname,
                    Void(pname=(self.nspname,aname[1:],), 
                        aname=aname,
                        oname='%s:%s' %(self.serialization_nsdict[self.nspname],aname[1:],),))
            elif not hasattr(pyclass,'literal'):
                # type definition
                setattr(self, aname,\
                    pyclass(pname=(self.nspname,aname[1:],), **self.serialization_nsdict))
            else:
                # element reference
                setattr(self, aname,\
                    pyclass(ns=self.nspname, **self.serialization_nsdict))
        elif not variable:
            if hasattr(typecode,'literal'):
                # element reference
                if not nil:
                    setattr(self, aname,\
                        typecode(ns=self.nspname, **self.serialization_nsdict))
                else:
                    raise EvaluateException, \
                        'must specify variable name to set a nil'
            else:
                raise EvaluateException, \
                    'must specify variable name(%s) for a type definition' %variable
        else:
            raise EvaluateException, \
                'variable name(%s) can not start with a double underscore' %variable

        if not nil and isinstance(getattr(self, aname), Struct):
            what = getattr(self, aname)
            what.declared = copy.copy(self.declared)

        
    def setDerivedTypeContents(self, extensions=None, restrictions=None):
        """For derived types set appropriate parameter and 
        """
        if extensions:
            ofwhat = list(self.ofwhat)
            if type(extensions) in _seqtypes:
                ofwhat += list(extensions)
            else:
                ofwhat.append(extensions)
        elif restrictions:
            if type(restrictions) in _seqtypes:
                ofwhat = restrictions
            else:
                ofwhat = (restrictions,)
        else:
            return
        self.ofwhat = tuple(ofwhat)
        self.lenofwhat = len(self.ofwhat)

    def parse(self, elt, ps):
        if self.type:
            targetNamespace, localName = self.checktype(elt, ps)
        else:
            targetNamespace, localName = elt.namespaceURI, elt.localName

        href = _find_href(elt)
        if href:
            if _children(elt):
                raise EvaluateException('Struct has content and HREF',
                        ps.Backtrace(elt))
            elt = ps.FindLocalHREF(href, elt)
        c = _child_elements(elt)
        count = len(c)
        if self.nilled(elt, ps): return None
        if count > self.lenofwhat and not self.hasextras:
            raise EvaluateException('Too many sub-elements', ps.Backtrace(elt))

        # Create the object.
        v = {}

        # Clone list of kids (we null it out as we process)
        c, crange = c[:], range(len(c))
        # Loop over all items we're expecting
        for i,what in [ (i, self.ofwhat[i]) for i in range(self.lenofwhat) ]:
            # Loop over all available kids
            for j,c_elt in [ (j, c[j]) for j in crange if c[j] ]:
                print_debug("WHATP(%s): " %what.aname, what)

                targetNamespace, localName = c_elt.namespaceURI, c_elt.localName
                tfix, tname = SplitQName(_find_type(c_elt))
                # XXX Not sure this is correct
                tNamespaceURI = targetNamespace
                my_elt = c_elt
                while tfix:
                    tNamespaceURI = my_elt.attributes.get((XMLNS.BASE, tfix))
                    if tNamespaceURI:
                        tNamespaceURI = tNamespaceURI._get_value()
                        print_debug("type namespace: ", tNamespaceURI)
                        break
                    my_elt = my_elt.parentNode
                my_elt = None
                del tfix

                if what.pname and what.name_match(c_elt):
                    # Parse value, and mark this one done.
                    try:
                        value = what.parse(c_elt, ps)
                    except EvaluateException, e:
                        if  not tname:
                            raise EvaluateException, 'instance not typed, can not look for derived type'
                        print_debug("Look for Derived Type(%s:%s) -- " %(tNamespaceURI, tname))
                        class_d = {}

                        if HasNamespaceContainer(tNamespaceURI):
                            container = GetNamespaceContainer(tNamespaceURI)
                            class_d = container.__dict__
                        for key,nsType in class_d.items():
                            if hasattr(nsType,'type') and hasattr(nsType,'schema'):
                                #Check against xsi:type -- namespace,name
                                kw = what.serialization_nsdict.copy()
                                kw['declared'] = copy.copy(self.declared)
                                if type(nsType.type) in _seqtypes and nsType.type == (tNamespaceURI,tname):
                                    what = nsType(pname=what.pname, **kw)
                                    break
                                elif nsType.type == tname and \
                                   nsType.schema == tNamespaceURI:
                                    what = nsType(pname=what.pname, **kw)
                                    break
                        else:
                            e.str = '%s.%s: %s' % \
                                    (self.pname or '?', what.aname or '?', e.str)
                            raise e
                        print_debug("WHATPP: ", what)
                        #Turn off self.lenofwhat check
                        what.hasextras = 1
                        value = what.parse(c_elt, ps)

                    if what.repeatable:
                        if v.has_key(what.aname):
                            v[what.aname].append(value)
                        else:
                            v[what.aname] = [value]
                        c[j] = None
                        continue
                    else:
                        v[what.aname] = value
                    c[j] = None
                    break
                elif isinstance(what, AnyElement):
                    if not ('#all' in what.namespaces) and \
                       not (tNamespaceURI in what.namespaces):
                        raise EvaluateException('namespace %s not in wildcard\'s purview' %tNamespaceURI,
                                ps.Backtrace(c_elt))

                    print_debug("AnyElement(%s,%s): \n\ttype = " %(tNamespaceURI, tname), self.type)

                    #Look for type, then element
                    ns_class = None
                    if self.type and HasNamespaceContainer(self.type[0]):
                        ns_class = GetNamespaceContainer(self.type[0])
                    elif HasNamespaceContainer(tNamespaceURI):
                        ns_class = GetNamespaceContainer(tNamespaceURI)
                    else:
                        raise EvaluateException, \
                             'AnyElement has no targetNamespace: %s' %tNamespaceURI

                    for key,nsType in ns_class.__dict__.items():
                        """First look for global type definition, then global 
                           element declaration.
                        """
                        if hasattr(nsType, 'type') and hasattr(nsType, 'schema') and not hasattr(nsType, 'literal'):
                            """ Global type definitions
                            """
                            print_debug("Global Type(%s): " %nsType.schema, nsType.type)
                            if (nsType.schema,nsType.type) == (tNamespaceURI, tname):
                                print_debug("FOUND GLOBAL TYPE: ", (tNamespaceURI, tname))
                                any_what = nsType(name=localName, ns=nsType.schema)
                                break
                            elif type(nsType.type) in _seqtypes:
                                if not hasattr(nsType, 'literal') and nsType.type == (tNamespaceURI, tname):
                                    print_debug("FOUND NEW GLOBAL TYPE: ", (tNamespaceURI, tname))
                                    any_what = nsType(name=localName, ns=nsType.schema)
                                    break
                                
                        elif hasattr(nsType,'literal') and hasattr(nsType,'schema'):
                            """ Global element declarations
                            """
                            print_debug("Global Element: (%s, %s)" %(nsType.schema, nsType.literal))
                            if nsType.literal == localName \
                                and nsType.schema == tNamespaceURI:
                                any_what = nsType()
                                break
                    else:
                        #XXX -- hard coded in variable aname here
                        any_what = XML(pname=localName, 
                                       aname='_%s'%localName, wrapped=0, 
                                       **{'copyit':1})
                        new_elt = c_elt
                        while new_elt.parentNode:
                            new_elt = new_elt.parentNode

                    print_debug("-->found: , ", any_what)
                    value = any_what.parse(c_elt, ps)
                    print_debug("-->parse: ", value)
                  
                    if what.repeatable:
                        if v.has_key(any_what.aname):
                            v[any_what.aname].append(value)
                        else:
                            v[any_what.aname] = [value]
                        c[j] = None
                        continue
                    else:
                        print_debug("any_what (%s): " %any_what.aname, any_what)
                        v[any_what.aname] = value
                    c[j] = None
                    break

                # No match; if it was supposed to be here, that's an error.
                if self.inorder and i == j:
                    raise EvaluateException('Out of order struct',
                            ps.Backtrace(c_elt))

            else:
                if not what.optional and not v.has_key(what.aname):
                    raise EvaluateException('Element "' + what.aname + \
                        '" missing from struct', ps.Backtrace(elt))
                if hasattr(what, 'default'):
                    v[what.aname] = what.default

        if not self.hasextras:
            extras = [c_elt for c_elt in c if c_elt]
            if len(extras):
                raise EvaluateException('Extra sub-elements (' + \
                        ','.join([c.nodeName for c in extras]) + ') found',
                        ps.Backtrace(elt))

        if not self.pyclass: return v
        print_debug("V: ", v)

        try:
            print_debug("pyclass: ", self.pyclass)
            kw = self.serialization_nsdict.copy()
            kw['declared'] = copy.copy(self.declared)
            if hasattr(self.pyclass, 'literal'):
                pyobj = self.pyclass(**kw)
            else:
                pyobj = self.pyclass(pname=self.pname, **kw)
        except Exception, e:
            raise TypeError("Constructing %s(%s): %s" % (self.pname, self.pyclass.__name__, str(e)))
        for key in v.keys():
            setattr(pyobj, key, v[key])

        #----------------------------------------
        # Get attributes from begin tag.
        #
        # class must define:
        #     attributeTypes -- class variable
        #     attributes -- instance variable
        #
        #----------------------------------------

        for a in _attrs(elt):
            attr_type = self.findAttributeType(\
                            attribute=a.localName,
                            namespace=a.namespaceURI,
                            what=pyobj)
            if attr_type:
                pyobj.attributes[a.localName] = a.nodeValue

        return pyobj

    def serialize(self, sw, pyobj, inline=None, name=None, attrtext='', **kw):
        print_debug("PYOBJ ", pyobj)
        if inline or self.inline:
            self.cb(sw, pyobj, **kw)
        else:
            objid = '%x' % id(pyobj)
            n = name or self.oname or ('E' + objid)
            print >>sw, '<%s%s href="#%s"/>' % (n, attrtext, objid)
            sw.AddCallback(self.cb, pyobj)

    def cb(self, sw, pyobj, name=None, **kw):
        declared = copy.copy(self.declared)
        if kw.has_key('declared'):
            declared += kw['declared']

        if not self.mutable and sw.Known(pyobj): return
        objid = '%x' % id(pyobj)

        #n = name or self.oname or ('E' + objid)
        if name or self.oname:
            if  self.serialization_nsdict.has_key(self.nspname):
                if self.serialization_nsdict[self.nspname]:
                    n = '%s:%s' %(self.serialization_nsdict[self.nspname], name or self.oname)
                    if self.nspname not in declared:
                        n += ' xmlns:%s="%s"' %(self.serialization_nsdict[self.nspname], self.nspname)
                        declared.append(self.nspname)
                else:
                    if self.nspname not in declared:
                        n = '%s xmlns="%s"' %(name or self.oname, self.nspname)
                        declared.append(self.nspname)
                    else:
                        n = '%s' %(name or self.oname)
            else:
                n = '%s xmlns="%s"' %(name or self.oname, self.nspname)
                self.serialization_nsdict[self.nspname] = None
                declared.append(self.nspname)
        else:
            n = name or self.oname or ('E' + objid)

        #---------------------------------------- 
        # Add type info to begin tag
        #----------------------------------------
        #if kw.get('typed', self.typed):
        if self.typed:
            if kw.get('type', self.type):
                ns, tag = kw.get('type', self.type)
                if self.serialization_nsdict.has_key(ns):
                   tag =  '%s:%s' %(self.serialization_nsdict[ns],tag)
                n += ' xsi:type="%s"' % tag

        #----------------------------------------
        # Add attributes to begin tag, assume all
        # attributes serialize with begin & end tags.
        #
        # class must define:
        #     attributeTypes -- class variable
        #     attributes -- instance variable
        #
        #----------------------------------------
        if hasattr(self, 'attributes'):
            for k,v in self.attributes.items():
                print_debug("Found Attribute %s = " %k, v)
                s = StringIO.StringIO()
                swa = SoapWriter(s,envelope=0)
                attr_type = self.findAttributeType(attribute=k)
                if v != None and isinstance(attr_type, TypeCode):
                    attr_type.serialize(swa, v)
                    xml_instance = s.getvalue()
                    endIdx = xml_instance.rfind('</')
                    begIdx = xml_instance[:endIdx].rfind('>')
                    attrValue = xml_instance[begIdx+1:endIdx]
                    if hasattr(self, 'attributeNS') and self.attributeNS.has_key(k):
                        n += ' %s:%s="%s"' %(self.serialization_nsdict[self.attributeNS[k]],k,attrValue)
                    else:
                        n += ' %s="%s"' %(k,attrValue)
                    s.truncate(0)

        if self.inline:
            print >>sw, '<%s>' % n
        else:
            print >>sw, '<%s id="%s">' % (n, objid)

        if self.pyclass:
            d = pyobj.__dict__
        else:
            d = pyobj
            if TypeCode.typechecks and type(d) != types.DictType:
                raise TypeError("Classless struct didn't get dictionary")

        for what in self.ofwhat:
            print_debug("WHAT: ", what)
            what_tc = what
            if isinstance(what, AnyElement):
                #AnyElement handling
                names = map(lambda f: f.aname, self.ofwhat)
                print_debug("D: ", d)
                for key in d.keys():
                    if d[key] and (not key in names) and (not key.lower() in names) \
                       and (key[0] == '_') and not (key[1] == '__'):
                        v = d[key]
                        print_debug("AnyElement FOUND: ", v)
                        #XXX SELF DESCRIBING 
                        #  OFWHAT MEANS INITIALIZED TYPECODE
                        if hasattr(v, 'ofwhat'):
                            what = v
                            print_debug("==>SELF-DESCRIBING")
                        elif hasattr(v, 'typecode'):
                            what = v.typecode 
                            print_debug("==>TYPECODE")
                        else:
                            what = XML(pname=what.pname, aname=what.aname, wrapped=0)
                        break
                else: 
                      continue
                #     raise EvaluateError, 'wildcard is empty'
            else:
                v = d.get(what.aname)
                if v is None:
                    v = d.get(what.aname.lower())
                if what.optional and v is None: 
                    continue

            if isinstance(v, what.__class__) and \
               v.__class__ != what.__class__ and \
               isinstance(v, TypeCode) and \
               hasattr(v.__class__, 'type'):
                #not hasattr(v, 'element'):
                # XXX Derived Types XXX
                # delcarations (elements) cannot be derived.
                # must be a definition.  Maybe this should change to 
                #
                print_debug("DERIVED TYPE(%s,%s): " %(v.pname, what.pname))
                print_debug("\tBase: ", v)
                print_debug("\tderv: ", what)
                what = v
            else:
                pass

            try:
                #------------------------------------------
                # serialize first check if instance is self-describing,
                # else use what of self.ofwhat
                #------------------------------------------
                #if what.repeatable and type(v) in _seqtypes:
                kw['declared'] = declared
                if what_tc.repeatable and type(v) in _seqtypes:
                    for v2 in v: 
                        if isinstance(v2, TypeCode):
                            v2.serialize(sw, v2, **kw)
                        what.serialize(sw, v2, **kw)
                elif isinstance(v, TypeCode):
                    print_debug('%s==>SELF-DESCRIBING' %v.pname)
                    v.serialize(sw, v, **kw)
                else:
                    print_debug("SERiaLize(%s): " %what.aname, what)
                    what.serialize(sw, v, **kw)
            except Exception, e:
                raise Exception('Serializing %s.%s, %s %s' %
                        (n, what.aname or '?', e.__class__.__name__, str(e)))


        # ignore the xmlns if it was explicitly stated
        #i = n.find('xmlns')
        i = n.find(' ')
        if i > 0:
            #print >>sw, '</%s>' % n[:i - 1]
            print >>sw, '</%s>' % n[:i]
        else:
            print >>sw, '</%s>' % n


class Array(TypeCode):
    '''An array.
        mutable -- object could change between multiple serializations
    '''

    def __init__(self, atype, ofwhat, pname=None, dimensions=1, fill=None,
    sparse=0, mutable=0, size=None, nooffset=0, undeclared=0,
    childnames=None, **kw):
        TypeCode.__init__(self, pname, **kw)
        self.dimensions = dimensions
        self.atype = atype
        if self.atype[-1] != ']': self.atype = self.atype + '[]'
        # Support multiple dimensions
        if self.dimensions != 1:
            raise TypeError("Only single-dimensioned arrays supported")
        self.fill = fill
        self.sparse = sparse
        if self.sparse: ofwhat.optional = 1
        self.mutable = mutable
        self.size = size
        self.nooffset = nooffset
        self.undeclared = undeclared
        self.childnames = childnames
        if self.size:
            t = type(self.size)
            if t in _inttypes:
                self.size = (self.size,)
            elif t in _seqtypes:
                self.size = tuple(self.size)
            elif TypeCode.typechecks:
                raise TypeError('Size must be integer or list, not ' + str(t))

        if TypeCode.typechecks:
            if type(atype) not in _stringtypes:
                raise TypeError("Array type must be a string.")
            t = type(ofwhat)
            if t != types.InstanceType:
                raise TypeError(
                    'Array ofwhat must be an instance, not ' + str(t))
            if not isinstance(ofwhat, TypeCode):
                raise TypeError(
                    'Array ofwhat outside the TypeCode hierarchy, ' +
                    str(ofwhat.__class__))
            if self.size:
                if len(self.size) != self.dimensions:
                    raise TypeError('Array dimension/size mismatch')
                for s in self.size:
                    if type(s) not in _inttypes:
                        raise TypeError('Array size "' + str(s) +
                                '" is not an integer.')
        self.ofwhat = ofwhat

    def parse_offset(self, elt, ps):
        o = _find_arrayoffset(elt)
        if not o: return 0
        if not _offset_pat.match(o):
            raise EvaluateException('Bad offset "' + o + '"',
                        ps.Backtrace(elt))
        return int(o[1:-1])

    def parse_position(self, elt, ps):
        o = _find_arrayposition(elt)
        if not o: return None
        if o.find(','):
            raise EvaluateException('Sorry, no multi-dimensional arrays',
                    ps.Backtrace(elt))
        if not _position_pat.match(o):
            raise EvaluateException('Bad array position "' + o + '"',
                    ps.Backtrace(elt))
        return int(o[-1:1])

    def parse(self, elt, ps):
        if not _find_arraytype(elt) and not self.undeclared:
            raise EvaluateException('Array expected', ps.Backtrace(elt))
        t = _find_type(elt)
        if t:
            pass # XXX should check the type, but parsing that is hairy.
        offset = self.parse_offset(elt, ps)
        v, vlen = [], 0
        if offset and not self.sparse:
            while vlen < offset:
                vlen += 1
                v.append(self.fill)
        for c in _child_elements(elt):
            #XXX
            #item = self.ofwhat.parse(c, ps)
            if self.ofwhat.name_match(c_elt):
                item = self.ofwhat.parse(c, ps)
            else:
                item = self.ofwhat.parse(c, ps)

            position = self.parse_position(c, ps) or offset
            if self.sparse:
                v.append((position, item))
            else:
                while offset < position:
                    offset += 1
                    v.append(self.fill)
                v.append(item)
            offset += 1
        return v

    def serialize(self, sw, pyobj, name=None, attrtext='', childnames=None,
    **kw):
        if not self.mutable and sw.Known(pyobj): return
        objid = '%x' % id(pyobj)
        n = name or self.oname or ('E' + objid)
        offsettext = ''
        if not self.sparse and not self.nooffset:
            offset, end = 0, len(pyobj)
            while offset < end and pyobj[offset] == self.fill:
                offset += 1
            if offset: offsettext = ' SOAP-ENC:offset="[%d]"' % offset
        if self.unique:
            idtext = ''
        else:
            idtext = ' id="%s"' % objid
        print >>sw, '<%s%s%s%s SOAP-ENC:arrayType="%s">' % \
                (n, attrtext, offsettext, idtext, self.atype)
        d = {}
        kn = childnames or self.childnames
        if kn:
            d['name'] = kn
        elif not self.ofwhat.aname:
            d['name'] = 'element'
        if not self.sparse:
            for e in pyobj[offset:]: self.ofwhat.serialize(sw, e, **d)
        else:
            position = 0
            for pos, v in pyobj:
                if pos != position:
                    d['attrtext'] = ' SOAP-ENC:position="[%d]"' % pos
                    position = pos
                else:
                    d['attrtext'] = ''
                self.ofwhat.serialize(sw, v, **d)
                position += 1
        print >>sw, '</%s>' % n


if __name__ == '__main__': print _copyright
