############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
#
###########################################################################

import urlparse
from xml.dom.ext.reader import PyExpat

""" Utilities.py


"""

DEBUG = 0
def print_debug(msg, arg=None):
    if DEBUG and arg:
        print "<UTIL>:  %s" %msg, arg
    elif DEBUG:
        print "<UTIL>:  %s" %msg


class HandleResolverDomAid:
    """This class returns reference wsdl defintions as dom objects.
    """
    post = '/ogsa/services/core/registry/handleResolver'
    handle_map = {
        '/ogsa/services/core/registry/ContainerRegistryService':'getNotificationServiceGroup',
        '/ogsa/services/core/NotificationSink':'getOGSINotificationSink',
        '/ogsa/services/sinks':'getOGSINotificationSubscription'
        }
    def __init__(self, handle, hostport):
        self.handle = urlparse.urlparse(handle)[2]
        self.addr, self.port = hostport

    def getReference(self, handle=None):
        if not self.addr or not self.port:
            raise RuntimeError, 'Must specify address and port'

        if handle:
             handle = urlparse.urlparse(handle)[2]
        else:
             handle = self.handle

        if self.handle_map.has_key(handle):
            wsdl = getattr(self, self.handle_map[handle])()
        else:
            raise RuntimeError, 'bad handle %s' %handle
        reader = PyExpat.Reader()
        return reader.fromString(wsdl)

    def getOGSINotificationSink(self):
        """OGSINotificationSink
        """
        wsdl = """
<definitions name="OGSINotificationSink" 
    targetNamespace="http://www.gridforum.org/namespaces/2003/03/OGSI/service" 
    xmlns="http://schemas.xmlsoap.org/wsdl/" 
    xmlns:ogsibinding="http://www.gridforum.org/namespaces/2003/03/OGSI/bindings" 
    xmlns:soap="http://schemas.xmlsoap.org/wsdl/soap/">
    <import location="http://%s:%d/schema/ogsi/ogsi_bindings.wsdl" 
            namespace="http://www.gridforum.org/namespaces/2003/03/OGSI/bindings"/>
    <service name="OGSINotificationSinkService" 
             xmlns:gsdl="http://ogsa.globus.org/">
         <gsdl:instanceOf handle="http://%s:%d/ogsa/services/sinks/10612" xmlns=""/>
         <port binding="ogsibinding:NotificationSinkSOAPBinding" 
               name="NotificationSinkPort">
             <soap:address location="http://%s:%d/ogsa/services/sinks/10612"/>
         </port>
    </service>
</definitions>""" %(self.addr,self.port,self.addr,self.port,self.addr,self.port)
        return wsdl

    def getNotificationServiceGroup(self):
        """NotificationServiceGroup:  This is ContainerRegistryService returns.
        """
        wsdl = """
<definitions name="NotificationServiceGroup" 
    targetNamespace="http://ogsi.gridforum.org/service" 
    xmlns="http://schemas.xmlsoap.org/wsdl/" 
    xmlns:notificationservicegroupbinding="http://ogsi.gridforum.org/bindings" 
    xmlns:soap="http://schemas.xmlsoap.org/wsdl/soap/">

    <import location="http://%s:%d/schema/ogsi/ogsi_notification_service_group_bindings.wsdl" 
        namespace="http://ogsi.gridforum.org/bindings"/>
    <service name="NotificationServiceGroupService" 
        xmlns:gsdl="http://ogsa.globus.org/">
        <gsdl:instanceOf 
            handle="http://%s:%d/ogsa/services/core/registry/ContainerRegistryService" 
            xmlns=""/>
        <port binding="notificationservicegroupbinding:NotificationServiceGroupSOAPBinding" 
            name="NotificationServiceGroupPort">
            <soap:address location="http://%s:%d/ogsa/services/core/registry/ContainerRegistryService"/>
        </port>
    </service>
</definitions>""" %(self.addr,self.port,self.addr,self.port,self.addr,self.port)
        return wsdl

    def getOGSINotificationSubscription(self):
        handle = "http://%s:%d/ogsa/services/ogsi/NotificationSubscriptionFactoryService/666"\
            %(self.addr, self.port)
        handle2 = "http://%s:%d/ogsa/services/instance" %(self.addr, self.port)
        location = "http://%s:%d/schema/ogsi/ogsi_bindings.wsdl" %(self.addr, self.port)
        wsdl = """
<definitions name="OGSINotificationSubscription" 
    targetNamespace="http://www.gridforum.org/namespaces/2003/03/OGSI/service" 
    xmlns="http://schemas.xmlsoap.org/wsdl/" 
    xmlns:ogsibinding="http://www.gridforum.org/namespaces/2003/03/OGSI/bindings" 
    xmlns:soap="http://schemas.xmlsoap.org/wsdl/soap/">

    <import location="%s" 
        namespace="http://www.gridforum.org/namespaces/2003/03/OGSI/bindings"/>
    <service name="OGSINotificationSubscription" xmlns:gsdl="http://ogsa.globus.org/">
        <gsdl:instanceOf handle="%s" xmlns=""/>
        <gsdl:instanceOf handle="%s" xmlns=""/>
        <port binding="ogsibinding:NotificationSubscriptionSOAPBinding" 
            name="NotificationSubscriptionPort">
            <soap:address location="%s"/>
        </port>
    </service>
</definitions>""" %(location, handle, handle2, handle)
        return wsdl
