############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import base64
from pyGridWare.ogsi.services.SecureContextEstablishment_services import \
    InitTokenExchangeInputMessageWrapper, ContinueTokenExchangeInputMessageWrapper

from pyGridWare import getLogger
logger = getLogger(__name__)


class Authenticator:
    """Establish a Security Context.
       secure-conversation, authorization=self, delegation=limited
       POST /ogsa/services/samples/counter/soap-secure/CounterFactoryService/authService
    """
    def __init__(self, context):
        """context -- GSSContext
        """
        self._context = context
        self._context_id = None

    def getContextID(self):
        return self._context_id

    def getContext(self):
        return self._context

    def setDelegation(self):
        self._context.setDelegation()

    def authenticate(self, port):
        """port -- SecureContextEstablishmentSOAPBindingSOAP
        """
        mechanism = "http://www.globus.org/2002/04/xmlenc#gssapi-gsi"
        token = self._context.initSecurityContext()
        initMsgIn = InitTokenExchangeInputMessageWrapper()
        initMsgIn._initContextToken.setAttribute_continue_needed(attribute=True)
        initMsgIn._initContextToken.setAttribute_mechanism_type(attribute=mechanism)
        initMsgIn._initContextToken._context_id = self._context.hashCode()
        initMsgIn._initContextToken._base64Token = base64.encodestring(token).strip()

        initMsgOut = port.initTokenExchange(initMsgIn)
        base64Token = initMsgOut._contextTokenOut._base64Token
        self._context_id = initMsgOut._contextTokenOut._context_id
        continue_needed = (initMsgOut._contextTokenOut.attributes['continue-needed']=='true')

        # continueTokenExchange
        # XXX Hack to get True set.
        while continue_needed:
            token = self._context.initSecurityContext(inToken=base64.decodestring(base64Token))
            base64Token = base64.encodestring(token).strip()
            cMsgIn = ContinueTokenExchangeInputMessageWrapper()
            cMsgIn._contextToken._context_id = self._context_id 
            cMsgIn._contextToken._base64Token = base64Token
            cMsgIn._contextToken.setAttribute_continue_needed(attribute=continue_needed)
            cMsgIn._contextToken.setAttribute_mechanism_type(attribute=mechanism)
            try:
                cMsgOut = port.continueTokenExchange(cMsgIn)
            except Exception, ex:
                print 'Exception: ', ex.args
                break
            else:
                continue_needed = (cMsgOut._contextTokenOut.attributes['continue-needed']=='true')
                base64Token = cMsgOut._contextTokenOut._base64Token
                self._context_id = cMsgOut._contextTokenOut._context_id


