############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import sys, urlparse
from pyGridWare.ogsi.services.SecureContextEstablishment_services import \
    SecureContextEstablishmentServiceLocator
from pyGridWare.ogsi.security.Authenticator import Authenticator
from pyGridWare.ogsi.security.GssXMLSignature import GssXMLSignature

from pyGridWare import getLogger
logger = getLogger(__name__)


class GenericHandler:
    """contexts -- dictionary of {context-id:context} 
    """
    #contexts = {}
    pass


class AuthenticationHandler(GenericHandler):
    def __init__(self):
        self.isDelegation = False

    def _getAuthServiceURL(self, url):
        """XXX hack 
        """
        return url + '/authService'

    def setDelegation(self):
        self.isDelegation = True

    def handleRequest(self, msgCtx):
        """msgCtx -- SOAPMessage
        """
        contextID = msgCtx.getContextID()
        if not contextID:
            locator = SecureContextEstablishmentServiceLocator()
            url = self._getAuthServiceURL(msgCtx.getURL())
            scheme,netloc,path,params,query,fragment = urlparse.urlparse(url)
            soap = locator.getSecureContextEstablishmentPortType(\
                       portAddress=url,
                       tracefile=sys.stdout, 
                       url=path)
            securityContext = msgCtx.getContext()
            authenticator = Authenticator(securityContext)
            if self.isDelegation:
                authenticator.setDelegation()
            authenticator.authenticate(port=soap)
            contextID = authenticator.getContextID()
            securityContext.setContextID(contextID)


class EncryptionHandler(GenericHandler):

    def __init__(self):
        self.useEncryption = False

    def setEncryption(self):
        self.useEncryption = True

    def handleRequest(self, msgCtx):
        if self.useEncryption:
            document = msgCtx.getMessageDocument()
            securityContext = msgCtx.getContext()
            contextID = securityContext.getContextID()

    def handleResponse(self, msgCtx):
       pass

class SignContentsHandler(GenericHandler):
    """adds the wsu:Id to the child element of SOAP-ENV:Body, marking
       it to be signed.
    """
    def handleRequest(self, msgCtx):
        #XXX HARDCODED digestSource, xmlns:wsu
        #
        from Ft.Xml import XPath
        from xml.ns import SOAP, XMLNS
        from pyGridWare.ogsi.security import WSU
        doc = msgCtx.getMessageDocument()
        xContext = XPath.Context.Context(doc, processorNss={'soapenv':SOAP.ENV, 'wsu':WSU.BASE})
        exp = XPath.Compile('soapenv:Envelope/soapenv:Body/*')
        nodes = exp.evaluate(xContext)
        if len(nodes) != 1:
            raise RuntimeError, 'SOAP-ENV:Body has multiple children'
        signNode = nodes[0]
        signNode.setAttributeNS(XMLNS.BASE, 'xmlns:wsu', WSU.BASE)
        signNode.setAttributeNS(namespaceURI=WSU.BASE, qualifiedName="wsu:Id", value="digestSource")


class SignatureHandler(GenericHandler):
    """signs requests, and validates responses
    """
    def handleRequest(self, msgCtx):
        document = msgCtx.getMessageDocument()
        signature = GssXMLSignature(doc=document)
        securityContext = msgCtx.getContext()
        contextID = securityContext.getContextID()
        SOAPData = signature.sign(context=securityContext, contextID=contextID)
        msgCtx.setMessage(SOAPData)

    def handleResponse(self, msgCtx):
        document = msgCtx.getMessageDocument()
        signature = GssXMLSignature(doc=document)
        securityContext = msgCtx.getContext()
        contextID = securityContext.getContextID()
        verify = signature.verify(context=securityContext, contextID=contextID)
        if not verify:
            raise RuntimeError, 'return message does not verify'

