############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
from xml.ns import SOAP, DSIG
from Ft.Xml import XPath
from pyGridWare.ogsi.security.ElementProxy import ElementProxy
from pyGridWare.ogsi.security import WSSE, WSU


class SignedInfo(ElementProxy):
    """
        <ds:SignedInfo>
            <ds:CanonicalizationMethod Algorithm="http://www.w3.org/TR/2001/REC-xml-c14n-20010315"></ds:CanonicalizationMethod>
            <ds:SignatureMethod Algorithm="http://www.globus.org/2002/04/xmlenc#gssapi-sign"></ds:SignatureMethod>
            <ds:Reference URI="#digestSource">
                <ds:DigestMethod Algorithm="http://www.w3.org/2000/09/xmldsig#sha1"></ds:DigestMethod>
                <ds:DigestValue>23ipYcsXfRxVglxNj1ogMz2TfaY=</ds:DigestValue>
            </ds:Reference>
        </ds:SignedInfo>
    """
    processorNss = {\
        'soapenv':SOAP.ENV,
        'wsse':WSSE.BASE,
        'wsu':WSU.BASE,
        'ds':DSIG.BASE,
        }
    DigestMethod = "http://www.w3.org/2000/09/xmldsig#sha1"
    CanonicalizationMethod = "http://www.w3.org/TR/2001/REC-xml-c14n-20010315"
    SignatureMethod = None
    siExp = XPath.Compile('soapenv:Envelope/soapenv:Header/wsse:Security/ds:Signature/ds:SignedInfo')

    def __init__(self, doc):
        ElementProxy.__init__(self, doc)
        self.setCanonicalizationURI(self.CanonicalizationMethod)
        self._sigMethod = self.SignatureMethod

    def getSignatureMethodURI(self):
        """Returns the Signature method URI  
        """
        return self._sigMethod

    def setSignatureMethodURI(SignatureMethodURI):
        """Sets the Signature method URI  
        """
        return

    def verify(self):
        """Method verify  
        """
        return

    def verifySignatureValue(self):
        """Method verifySignatureValue
        """
        return

