############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
#
###########################################################################

import ZSI
from ZSI import _attrs, _child_elements, _children
from StringIO import StringIO
from xml.dom.ext.reader import PyExpat
from pyGridWare.ogsi.services.properties_detail_xsd_services_types import \
    ogsa_globus_org_types_properties as ns4

#from pyGridWare.ogsi.services.OGSIService_services import \
#    FindByHandleOutputMessage, ns1

from pyGridWare.ogsi.services.OGSI_services import \
    FindByHandleOutputMessageWrapper, ns1

from pyGridWare.ogsi.services.NotificationFactoryService_services import \
    FindServiceDataOutputMessageWrapper,\
    SubscribeOutputMessageWrapper

from pyGridWare.ogsi.services.NotificationFactoryService_services import \
    ns1 as ns2

from pyGridWare.ogsi.services.GWSDL_types import \
    www_gridforum_org_namespaces_2003_03_serviceData as ns3

#from pyGridWare.ogsi.services.NotificationFactoryService_services import \
#    www_gridforum_org_namespaces_2003_03_serviceData as ns3

#from pyGridWare.ogsi.services.properties_detail_xsd_services_types import \
#    ogsa_globus_org_types_properties as ns4
    

""" NotificationModule

     This is one of the modules passed to dispatch for the stand-alone server.
    This module should define GridService interactions.

    HandleResolver 
        findByHandle -- sort of works
  
    ServiceData
        findServiceData -- need to store and retrieve these values

    Notification 
        subscribe
        deliverConfimation

"""

ADDR = None
PORT = None
DEBUG = 5
def print_debug(msg, arg=None):
    if DEBUG and arg:
        print "<MOD>:  %s" %msg, arg
    elif DEBUG:
        print "<MOD>:  %s" %msg


###############################################################################
#      CallBacks
#
###############################################################################
def requestTerminationBefore(*args):
    #RequestTerminationBeforeInputMessageWrapper()
    #RequestTerminationBeforeOutputMessageWrapper()
    print_debug("requestTerminationBefore(*args):", args)
    return None


def subscribe(*args):
    #SubscribeInputMessage
    #SubscribeOutputMessageWrapper()
    print_debug("def subscribe(*args):", args)
    return getSubscribeOutputMessage(args[0])


def requestTerminationAfter(*args):
    #RequestTerminationAfterInputMessage
    #RequestTerminationAfterOutputMessageWrapper()
    print_debug("def requestTerminationAfter(*args):", args)
    return None


def findServiceData(*args):
    """
  REQUEST ---------------------------------------
 <soapenv:Body>
  <findServiceData xmlns="http://www.gridforum.org/namespaces/2003/03/OGSI">
   <queryExpression>
    <queryByServiceDataNames xsi:type="ns1:QNamesType" xmlns:ns1="http://www.gridforum.org/namespaces/2003/03/OGSI">
     <name>subscribeExtensibility</name>
    </queryByServiceDataNames>
   </queryExpression>
  </findServiceData>
 </soapenv:Body>

  RESPONSE --------------------------------------
 <soapenv:Body>
  <findServiceDataResponse xmlns="http://www.gridforum.org/namespaces/2003/03/OGSI">
   <result xsi:type="ns1:ExtensibilityType" 
           xmlns:ns1="http://www.gridforum.org/namespaces/2003/03/OGSI">
    <ns2:serviceDataValues xmlns:ns2="http://www.gridforum.org/namespaces/2003/03/serviceData">
     <subscribeExtensibility inputElement="ogsi:subscribeByServiceDataNames" 
                             xmlns:ogsi="http://www.gridforum.org/namespaces/2003/03/OGSI"/>
    </ns2:serviceDataValues>
   </result>
  </findServiceDataResponse>
 </soapenv:Body>
</soapenv:Envelope>

    """
    msg = args[0]
    print_debug("----- findServiceData(*args):", args)
    print_debug('\t_queryExpression = ', msg._queryExpression)
    print_debug('\t\t_queryByServiceDataNames = ', msg._queryExpression._queryByServiceDataNames)
    print_debug('\t\t\t_name = ', msg._queryExpression._queryByServiceDataNames._name)

    resp = getFindServiceDataOutputMessage(msg)
    return resp


def setServiceData(*args):
    #SetServiceDataInputMessage
    #SetServiceDataOutputMessageWrapper()
    print_debug("def setServiceData(*args):", args)
    return None


def destroy(*args):
    #DestroyInputMessage
    #DestroyOutputMessageWrapper
    print_debug("def destroy(*args):", args)
    return None


def deliverNotification(*args):
    request = args[0]
    print_debug(" ---- deliverNotification")
    print_debug("message: ", request.message)
    print_debug("\tserviceDataValues -- ", request.message.serviceDataValues)
    print_debug("\t\tterminationTime -- ", request.message.serviceDataValues._terminationTime)
    for k,v in request.message.serviceDataValues._terminationTime.attributes_dict.items():  
        print_debug("\t\t\t%s  -- %s" %(k,v))

    return DeliverNotificationResponse()


def findByHandle(*args):
    # Only one possibility
    # GSH --> GSR
    msg = args[0]
    print_debug('findByHandleXXX: ', msg)
    return getFindByHandleResponse(msg)



###############################################################################
# HELPER FUNCTIONS
# --WSDL Definition DOM
###############################################################################
import urlparse
class HandleResolverDomAid:
    """This class returns reference wsdl defintions as dom objects.
    """
    post = '/ogsa/services/core/registry/handleResolver'
    handle_map = {
        '/ogsa/services/core/registry/ContainerRegistryService':'getNotificationServiceGroup',
        '/ogsa/services/core/NotificationSink':'getOGSINotificationSink',
        '/ogsa/services/sinks':'getOGSINotificationSubscription'
        }
    def __init__(self, handle, hostport):
        self.handle = urlparse.urlparse(handle)[2]
        self.addr, self.port = hostport

    def getReference(self, handle=None):
        if not self.addr or not self.port:
            raise RuntimeError, 'Must specify address and port'

        if handle:
             handle = urlparse.urlparse(handle)[2]
        else:
             handle = self.handle

        if self.handle_map.has_key(handle):
            wsdl = getattr(self, self.handle_map[handle])()
        else:
            raise RuntimeError, 'bad handle %s' %handle
        reader = PyExpat.Reader()
        return reader.fromString(wsdl)

    def getOGSINotificationSink(self):
        """OGSINotificationSink
        """
        wsdl = """
<definitions name="OGSINotificationSink" 
    targetNamespace="http://www.gridforum.org/namespaces/2003/03/OGSI/service" 
    xmlns="http://schemas.xmlsoap.org/wsdl/" 
    xmlns:ogsibinding="http://www.gridforum.org/namespaces/2003/03/OGSI/bindings" 
    xmlns:soap="http://schemas.xmlsoap.org/wsdl/soap/">
    <import location="http://%s:%d/schema/ogsi/ogsi_bindings.wsdl" 
            namespace="http://www.gridforum.org/namespaces/2003/03/OGSI/bindings"/>
    <service name="OGSINotificationSinkService" 
             xmlns:gsdl="http://ogsa.globus.org/">
         <gsdl:instanceOf handle="http://%s:%d/ogsa/services/sinks/10612" xmlns=""/>
         <port binding="ogsibinding:NotificationSinkSOAPBinding" 
               name="NotificationSinkPort">
             <soap:address location="http://%s:%d/ogsa/services/sinks/10612"/>
         </port>
    </service>
</definitions>""" %(self.addr,self.port,self.addr,self.port,self.addr,self.port)
        return wsdl

    def getNotificationServiceGroup(self):
        """NotificationServiceGroup:  This is ContainerRegistryService returns.
        """
        wsdl = """
<definitions name="NotificationServiceGroup" 
    targetNamespace="http://ogsi.gridforum.org/service" 
    xmlns="http://schemas.xmlsoap.org/wsdl/" 
    xmlns:notificationservicegroupbinding="http://ogsi.gridforum.org/bindings" 
    xmlns:soap="http://schemas.xmlsoap.org/wsdl/soap/">

    <import location="http://%s:%d/schema/ogsi/ogsi_notification_service_group_bindings.wsdl" 
        namespace="http://ogsi.gridforum.org/bindings"/>
    <service name="NotificationServiceGroupService" 
        xmlns:gsdl="http://ogsa.globus.org/">
        <gsdl:instanceOf 
            handle="http://%s:%d/ogsa/services/core/registry/ContainerRegistryService" 
            xmlns=""/>
        <port binding="notificationservicegroupbinding:NotificationServiceGroupSOAPBinding" 
            name="NotificationServiceGroupPort">
            <soap:address location="http://%s:%d/ogsa/services/core/registry/ContainerRegistryService"/>
        </port>
    </service>
</definitions>""" %(self.addr,self.port,self.addr,self.port,self.addr,self.port)
        return wsdl

    def getOGSINotificationSubscription(self):
        handle = "http://%s:%d/ogsa/services/ogsi/NotificationSubscriptionFactoryService/666"\
            %(self.addr, self.port)
        handle2 = "http://%s:%d/ogsa/services/instance" %(self.addr, self.port)
        location = "http://%s:%d/schema/ogsi/ogsi_bindings.wsdl" %(self.addr, self.port)
        wsdl = """
<definitions name="OGSINotificationSubscription" 
    targetNamespace="http://www.gridforum.org/namespaces/2003/03/OGSI/service" 
    xmlns="http://schemas.xmlsoap.org/wsdl/" 
    xmlns:ogsibinding="http://www.gridforum.org/namespaces/2003/03/OGSI/bindings" 
    xmlns:soap="http://schemas.xmlsoap.org/wsdl/soap/">

    <import location="%s" 
        namespace="http://www.gridforum.org/namespaces/2003/03/OGSI/bindings"/>
    <service name="OGSINotificationSubscription" xmlns:gsdl="http://ogsa.globus.org/">
        <gsdl:instanceOf handle="%s" xmlns=""/>
        <gsdl:instanceOf handle="%s" xmlns=""/>
        <port binding="ogsibinding:NotificationSubscriptionSOAPBinding" 
            name="NotificationSubscriptionPort">
            <soap:address location="%s"/>
        </port>
    </service>
</definitions>""" %(location, handle, handle2, handle)
        return wsdl



def getFindByHandleResponse(msg):
    """GSR Producer
    """
    if msg:
        urls = msg._handleSet._handle
    else:
        urls = '/ogsa/services/core/registry/ContainerRegistryService'
        #url = '/ogsa/services/core/NotificationSink'

    resolver = HandleResolverDomAid(handle=urls[0], hostport=(ADDR,PORT))
    pyobj = FindByHandleOutputMessageWrapper()
    pyobj._locator = ns1.LocatorType_Def(name="locator")
    pyobj._locator._reference = ns1.WSDLReferenceType_Def(name='reference')
    try:
        pyobj._locator._reference._dom = resolver.getReference()
    except Exception, ex:
        print ex
        return
    """
    pyobj.__class__.typecode.ofwhat[0].oname = '%s  %s  %s' %('locator',
        'xmlns:ns1="http://www.gridforum.org/namespaces/2003/03/OGSI"',
        'xsi:type="ns1:LocatorType"')
    """

    #COPY IT TO DERIVED TYPE
    """
    pyobj.__class__.typecode.ofwhat[0].ofwhat[1].oname = '%s  %s  %s' %('reference',
        'xmlns:ns1="http://www.gridforum.org/namespaces/2003/03/OGSI"',
        'xsi:type="ns1:WSDLReferenceType"')
    pyobj._locator._reference.oname = pyobj.__class__.typecode.ofwhat[0].ofwhat[1].oname
    """


    return pyobj


def getFindServiceDataOutputMessage(msg):
    """
  RESPONSE --------------------------------------
  --------------------------------------
    <element name="findServiceDataResponse">
      <complexType>
        <sequence>
          <element name="result" type="ogsi:ExtensibilityType"/>
        </sequence>
      </complexType>
    </element>

  <complexType name="ServiceDataValuesElementType">
    <sequence>
      <element name="serviceDataValues" type="tns:ServiceDataValuesType"/>
    </sequence>
  </complexType>
  --------------------------------------
  <findServiceDataResponse xmlns="http://www.gridforum.org/namespaces/2003/03/OGSI">
   <result xsi:type="ns1:ExtensibilityType" 
           xmlns:ns1="http://www.gridforum.org/namespaces/2003/03/OGSI">
    <ns2:serviceDataValues xmlns:ns2="http://www.gridforum.org/namespaces/2003/03/serviceData">

   1.
     <subscribeExtensibility inputElement="ogsi:subscribeByServiceDataNames" 
                             xmlns:ogsi="http://www.gridforum.org/namespaces/2003/03/OGSI"/>

   2.
     <entry xsi:type="ns1:EntryType">
      <serviceGroupEntryLocator xsi:nil="true"/>
      <memberServiceLocator xsi:type="ns1:LocatorType">
       <handle xsi:type="ns1:HandleType">http://192.168.1.2:8081/ogsa/services/samples/counter/notification/CounterFactoryService</handle>
      </memberServiceLocator>
      <content xsi:type="ns1:EntryContentType">
       <ns13:propertiesDetail xsi:type="ns13:PropertiesDetailType" xmlns:ns13="http://ogsa.globus.org/types/properties">
        <ns13:name>samples/counter/notification/CounterFactoryService</ns13:name>
        <ns13:portType></ns13:portType>
        <ns13:state xsi:type="ns13:StateType">INACTIVE</ns13:state>
        <ns13:propertiesLifecycle xsi:type="ns13:LifeCycleType">TRANSIENT</ns13:propertiesLifecycle>
        <ns13:deploymentLifecycle xsi:type="ns13:LifeCycleType">TRANSIENT</ns13:deploymentLifecycle>
        <ns13:activateOnStartup>false</ns13:activateOnStartup>
        <ns13:handle>http://192.168.1.2:8081/ogsa/services/samples/counter/notification/CounterFactoryService</ns13:handle>
       </ns13:propertiesDetail>
      </content>
     </entry>
     ....

    </ns2:serviceDataValues>
   </result>
  </findServiceDataResponse>


  --------------------------------------
    """
    response = FindServiceDataOutputMessageWrapper()
    response._result._serviceDataValues = ns3.ServiceDataValuesElement_Dec()
    if msg:
        #---------------------------------------------------
        #QNamesType is repeatable
        #  I have no idea what to do if it repeats, so just break
        #---------------------------------------------------
        names = ['entry']
        if hasattr(msg,'_queryExpression'):
            names = msg._queryExpression._queryByServiceDataNames._name

        for name in names:
            if name == 'subscribeExtensibility':
                response._result._serviceDataValues._subscribeExtensibility  = \
                    ns3.subscribeExtensibility_Def(name='subscribeExtensibility',
                                                   repeatable=1,
                                                   optional=1)
                break
            elif name == 'entry':
                response._result._serviceDataValues._entry = ns2.EntryType_Def(name='entry')
                response._result._serviceDataValues._entry._serviceGroupEntryLocator = ZSI.TC.Void('serviceGroupEntryLocator')
                #response._result._serviceDataValues._entry._serviceGroupEntryLocator = \
                #    ns2.LocatorType(name="serviceGroupEntryLocator", ns=ns, optional=1)
                response._result._serviceDataValues._entry._memberServiceLocator = \
                    ns2.LocatorType_Def(name="memberServiceLocator")
                response._result._serviceDataValues._entry._memberServiceLocator._handle = \
                    "http://%s:%d/ogsa/services/samples/counter/notification/CounterFactoryService" \
                    %(ADDR,PORT)
                    
                response._result._serviceDataValues._entry._content = \
                    ns2.EntryContentType_Def(name="content")
                response._result._serviceDataValues._entry._content._propertiesDetail = \
                    ns4.PropertiesDetailType_Def(name='propertiesDetail',typed=1)

                properties = response._result._serviceDataValues._entry._content._propertiesDetail
                properties._name = "samples/counter/notification/CounterFactoryService"
                #properties._portType = ZSI.TC.Struct(None, 'portType')
                properties._portType = Holder()
                properties._state = "INACTIVE"
                properties._propertiesLifecycle = "TRANSIENT"
                properties._deploymentLifecycle = "TRANSIENT"
                properties._activateOnStartup = 0
                properties._handle = \
                    "http://%s:%d/ogsa/services/samples/counter/notification/CounterFactoryService"\
                    %(ADDR, PORT)
          
                break
            else: 
                raise RuntimeError, 'Bad Service Data Request'

    return response

class Holder: pass

import time
def getSubscribeOutputMessage(request):
    """------------------------------------------
   ----------------------------------------------
  <subscribe xmlns="http://www.gridforum.org/namespaces/2003/03/OGSI">
   <subscriptionExpression>
    <subscribeByServiceDataNames xsi:type="ns1:SubscribeByNameType" 
        xmlns:ns1="http://www.gridforum.org/namespaces/2003/03/OGSI">
     <name>entry</name>
    </subscribeByServiceDataNames>
   </subscriptionExpression>
   <sink>
    <handle>http://192.168.1.2:32965/ogsa/services/sinks/1066769825928-192.168.1.2-32965-13390902-11320634</handle>
   </sink>
   <expirationTime>2003-10-21T21:57:07.247Z</expirationTime>
  </subscribe>
   ----------------------------------------------
   ----------------------------------------------
  <subscribeResponse xmlns="http://www.gridforum.org/namespaces/2003/03/OGSI">
   <subscriptionInstanceLocator 
       xsi:type="ns1:LocatorType" 
       xmlns:ns1="http://www.gridforum.org/namespaces/2003/03/OGSI">
      <reference xsi:type="ns1:WSDLReferenceType">
        <definitions name="OGSINotificationSubscription" 
            targetNamespace="http://www.gridforum.org/namespaces/2003/03/OGSI/service" 
            xmlns="http://schemas.xmlsoap.org/wsdl/" 
            xmlns:ogsibinding="http://www.gridforum.org/namespaces/2003/03/OGSI/bindings" 
            xmlns:soap="http://schemas.xmlsoap.org/wsdl/soap/">
          <import location="http://192.168.1.2:8081/schema/ogsi/ogsi_bindings.wsdl" 
                  namespace="http://www.gridforum.org/namespaces/2003/03/OGSI/bindings"/>
          <service name="OGSINotificationSubscription" 
                   xmlns:gsdl="http://ogsa.globus.org/">
              <gsdl:instanceOf handle="http://192.168.1.2:8081/ogsa/services/ogsi/NotificationSubscriptionFactoryService/hash-26675936-1066769829203" xmlns=""/>
              <gsdl:instanceOf handle="http://192.168.1.2:8081/ogsa/services/instance" xmlns=""/>
              <port binding="ogsibinding:NotificationSubscriptionSOAPBinding" 
                    name="NotificationSubscriptionPort">
                  <soap:address location="http://192.168.1.2:8081/ogsa/services/ogsi/NotificationSubscriptionFactoryService/hash-26675936-1066769829203"/>
              </port>
          </service>
        </definitions>    
    </reference>
   </subscriptionInstanceLocator>
   <ns2:currentTerminationTime xsi:type="ns2:TerminationTimeType" 
       ns2:after="2003-10-21T21:57:07.247Z" 
       ns2:before="2003-10-21T21:57:07.247Z" 
       ns2:timestamp="2003-10-21T20:57:09.237Z" 
       xmlns:ns2="http://www.gridforum.org/namespaces/2003/03/OGSI"/>
  </subscribeResponse>
   ----------------------------------------------
    """
    #request._subscriptionExpression
    #request._subscriptionExpression._subscribeByServiceDataNames
    #request._subscriptionExpression._subscribeByServiceDataNames._name
    #request._sink
    #request._sink._handle
    #request._expirationTime

    url = '/ogsa/services/sinks'
    pyobj = SubscribeOutputMessageWrapper()
    pyobj._currentTerminationTime = Holder()
    #pyobj._currentTerminationTime = time.gmtime(time.time()+10000)

    """------------------------------------------
       Normal _subscriptionInstanceLocator --> class typecode
       ------------------------------------------
    pyobj.__class__.typecode.ofwhat[0].oname = '%s %s %s'\
        %(pyobj._subscriptionInstanceLocator.oname,\
          'xsi:type="ns1:LocatorType"',\
          'xmlns:ns1="http://www.gridforum.org/namespaces/2003/03/OGSI"')
    """

    """------------------------------------------
       DerivedType _reference --> instance's typecode
       ------------------------------------------
    """
    pyobj._subscriptionInstanceLocator._reference = \
        ns2.WSDLReferenceType_Def(name='reference')
    #pyobj._subscriptionInstanceLocator._reference.oname = '%s %s %s' \
    #    %(pyobj._subscriptionInstanceLocator._reference.oname,\
    #      'xmlns:ns1="http://www.gridforum.org/namespaces/2003/03/OGSI"',\
    #      'xsi:type="ns1:WSDLReferenceType"')


    resolver = HandleResolverDomAid(handle=url, hostport=(ADDR,PORT))
    try:
        pyobj._subscriptionInstanceLocator._reference._dom = resolver.getReference()
    except Exception, ex:
        raise

    #pyobj.__class__.typecode.ofwhat[0].oname = '%s  %s  %s' %('locator',
    #    'xmlns:ns1="http://www.gridforum.org/namespaces/2003/03/OGSI"',
    #    'xsi:type="ns1:LocatorType"')

    #COPY IT TO DERIVED TYPE
    #pyobj.__class__.typecode.ofwhat[0].ofwhat[1].oname = '%s  %s  %s' %('reference',
    #    'xmlns:ns1="http://www.gridforum.org/namespaces/2003/03/OGSI"',
    #    'xsi:type="ns1:WSDLReferenceType"')
    #pyobj._locator._reference.oname = pyobj.__class__.typecode.ofwhat[0].ofwhat[1].oname

    return pyobj


###############################################################################
#  BELOW -- REMOVE THIS STUFF
###############################################################################
def main():
    import socket
    global ADDR, PORT
    #ADDR = socket.gethostbyname(socket.gethostname())
    ADDR = '192.168.40.101'
    PORT = 8080
    #pyobj = getFindServiceDataOutputMessage(1)
    pyobj = getFindByHandleResponse(None)
    #pyobj = getSubscribeOutputMessage(None)
    print "MSG: ", pyobj
    s = StringIO()
    sw = ZSI.SoapWriter(s)
    sw.serialize(pyobj)
    print "RESULT -----\n", s.getvalue()

if __name__ == '__main__':
    main()

