############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
#
###########################################################################
import urlparse, time
from ZSI import _seqtypes

from pyGridWare.ogsi.services.OGSI_services import \
    DestroyOutputMessageWrapper, FindServiceDataOutputMessageWrapper,\
    RequestTerminationAfterOutputMessageWrapper,\
    RequestTerminationBeforeOutputMessageWrapper, SetServiceDataOutputMessageWrapper,\
    SubscribeOutputMessageWrapper, NotificationSinkSOAPBindingSOAP, ns1

from pyGridWare.ogsi.services.GWSDL_types import \
    www_gridforum_org_namespaces_2003_03_serviceData as ns3

from properties_detail_xsd_services_types import ns1 as ns4

from pyGridWare.ogsi.services.GridService import \
    GridServiceInterface, PortType, GridService_portType

from pyGridWare.ogsi.services import OGSINotificationSubscription

from pyGridWare import getLogger
logger = getLogger(__name__)
    
""" NotificationServiceGroupService

     This is one of the modules passed to dispatch for the stand-alone server.
    This module should define GridService interactions.

    classes: 


"""
class Holder: pass

class NotificationSource_portType(GridService_portType):
    def setServiceData(self, msg):
        raise NotImplementedError, 'portType is a interface'

class NotificationSource_gPortType(PortType, NotificationSource_portType, GridService_gPortType):
    """
    <gwsdl:portType name="NotificationSource" extends="ogsi:GridService">
      <operation name="subscribe">
        <input message="ogsi:SubscribeInputMessage"/>
        <output message="ogsi:SubscribeOutputMessage"/>
        <fault name="ExtensibilityNotSupportedFault" 
            message="ogsi:ExtensibilityNotSupportedFaultMessage"/>
        <fault name="ExtensibilityTypeFault" 
            message="ogsi:ExtensibilityTypeFaultMessage"/>
        <fault name="TargetInvalidFault" 
            message="ogsi:TargetInvalidFaultMessage"/>
        <fault name="Fault" message="ogsi:FaultMessage"/>
      </operation>  

      <sd:staticServiceDataValues>
          <ogsi:subscribeExtensibility inputElement="ogsi:subscribeByServiceDataNames"/>
      </sd:staticServiceDataValues>
    """
    #extends..
    serviceDataTypes = GridService_gPortType.serviceDataTypes.copy()
    class notifiableServiceDataName(PortType.ServiceData):
        name="notifiableServiceDataName"
        type=TCmodified.QName
        minOccurs=0
        maxOccurs="unbounded"
        mutability="mutable"
        modifiable=False
        nillable=False
    serviceDataTypes[notifiableServiceDataName.name] = notifiableServiceDataName

    class subscribeExtensibility(PortType.ServiceData):
        """<complexType name="OperationExtensibilityType">
             <attribute name="inputElement" type="QName" use="optional"/>
           </complexType>
        """
        name="subscribeExtensibility"
        type=ns1.OperationExtensibilityType
        minOccurs=1
        maxOccurs="unbounded"
        mutability="static"
        modifiability=False
        nillable=False
        def __init__(self):
            sd = self.addInstance()
            sd.setAttribute_inputElement(('http://www.gridforum.org/namespaces/2003/03/OGSI','subscribeByServiceDataNames'))

    def __init__(self, addr, port, post, reference):
        PortType.__init__(self)
        NotificationSource_portType.__init__(self, addr, port, post, reference)

    def subscribeExtensibility_SDE(self, pyobj):
        """<subscribeExtensibility 
               inputElement="ogsi:subscribeByServiceDataNames" 
               xmlns:ogsi="http://www.gridforum.org/namespaces/2003/03/OGSI"/>
        """
        logger.info('NotificationSource_gPortType.subscribeExtensibility')
        return self.staticServiceData[NotificationSource_gPortType.subscribeExtensibility.name].getInstances()

    def notifiableServiceDataName_SDE(self, pyobj):
        logger.info('NotificationSource_gPortType.notifiableServiceDataName')
        return self.staticServiceData[NotificationSource_gPortType.subscribeExtensibility.name].getInstances()


class ServiceGroup_portType:
    pass

class ServiceGroup_gPortType(PortType, ServiceGroup_portType, GridService_gPortType):
    """
    <gwsdl:portType name="ServiceGroup" extends="ogsi:GridService">
    """
    class membershipContentRule(PortType.ServiceData):
        name="membershipContentRule"
        type=ns1.MembershipContentRuleType_Def
        minOccurs="1"
        maxOccurs="unbounded"
        mutability="constant" 
        modifiable="false"
        nillable="false"

    class name(PortType.ServiceData):
        name="entry"
        type=ns1.EntryType_Def
        minOccurs="0"
        maxOccurs="unbounded"
        mutability="mutable"
        modifiable="false"

    serviceDataTypes = GridService_gPortType.serviceDataTypes.copy()
    serviceDataTypes[membershipContentRule.name] = membershipContentRule
    serviceDataTypes[name.name] = name

    def __init__(self, addr, port, post, reference):
        PortType.__init__(self)
        GridService_gPortType.__init__(self, addr, port, post, reference)

    def membershipContentRule_SDE(self, pyobj):
        logger.info('ServiceGroup_gPortType.membershipContentRule')
        return self.serviceData[ServiceGroup_gPortType.membershipContentRule.name].getInstances()

    def name_SDE(self, pyobj):
        logger.info('ServiceGroup_gPortType.name')
        return self.serviceData[ServiceGroup_gPortType.name.name].getInstances()


class NotificationServiceGroup_portType:
    pass

class NotificationServiceGroup_gPortType(PortType, NotificationServiceGroup_portType, NotificationSource_gPortType, ServiceGroup_gPortType):
    #EXTENDS
    serviceDataTypes = ServiceGroup_gPortType.serviceDataTypes.copy()
    serviceDataTypes.update(NotificationSource_gPortType.serviceDataTypes)

    def __init__(self, addr, port, post, reference):
        PortType.__init__(self)
        NotificationSource_gPortType.__init__(self, addr, port, post, reference)
        ServiceGroup_gPortType.__init__(self, addr, port, post, reference)


class NotificationServiceGroupSOAPBinding(NotificationServiceGroup_gPortType):
    """
    <binding name="NotificationServiceGroupSOAPBinding" 
             type="notificationservicegroup:NotificationServiceGroup">
        <soap:binding style="document" transport="http://schemas.xmlsoap.org/soap/http"/>
            <operation name="subscribe">
            <operation name="setServiceData">
            <operation name="destroy">
            <operation name="requestTerminationAfter">
            <operation name="findServiceData">
            <operation name="requestTerminationBefore">
    """
    soapAction = {\
        'http://ogsi.gridforum.org#subscribe':'subscribe',
        'http://ogsi.gridforum.org#setServiceData':'setServiceData',
        'http://ogsi.gridforum.org#destroy':'destroy',
        'http://ogsi.gridforum.org#requestTerminationAfter':'requestTerminationAfter',
        'http://ogsi.gridforum.org#findServiceData':'findServiceData',
        'http://ogsi.gridforum.org#requestTerminationBefore':'requestTerminationBefore',
        }

    def subscribe(self, msg):
        logger.info('NotificationServiceGroupSOAPBinding.subscribe')
        #SubscribeInputMessage
        msg._subscriptionExpression
        msg._subscriptionExpression._subscribeByServiceDataNames
        msg._subscriptionExpression._subscribeByServiceDataNames._name
        msg._sink
        msg._sink._handle
        msg._expirationTime

        #SubscribeOutputMessage
        pyobj = SubscribeOutputMessageWrapper()
        return pyobj

    def findServiceData(self, msg):
        logger.info('NotificationServiceGroupSOAPBinding.findServiceData')
        pyobj = FindServiceDataOutputMessageWrapper()
        return pyobj

    def setServiceData(self, msg):
        logger.info('NotificationServiceGroupSOAPBinding.setServiceData')
        pyobj = SetServiceDataOutputMessageWrapper()
        return pyobj

    def destroy(self, msg):
        logger.info('NotificationServiceGroupSOAPBinding.destroy')
        pyobj = DestroyOutputMessageWrapper()
        return pyobj

    def requestTerminationBefore(self, msg):
        logger.info('NotificationServiceGroupSOAPBinding.requestTerminationBefore')
        pyobj = RequestTerminationBeforeOutputMessageWrapper()
        return pyobj

    def requestTerminationAfter(self, msg):
        logger.info('NotificationServiceGroupSOAPBinding.requestTerminationAfter')
        pyobj = RequestTerminationAfterOutputMessageWrapper()
        return pyobj


class NotificationServiceInterface(GridServiceInterface):
    """
  <subscribe xmlns="http://www.gridforum.org/namespaces/2003/03/OGSI">
   <subscriptionExpression>
    <subscribeByServiceDataNames xsi:type="ns1:SubscribeByNameType" xmlns:ns1="http://www.gridforum.org/namespaces/2003/03/OGSI">
     <name>entry</name>
    </subscribeByServiceDataNames>
   </subscriptionExpression>
   <sink>
    <handle>http://192.168.1.2:32965/ogsa/services/sinks/1066769825928-192.168.1.2-32965-13390902-11320634</handle>
   </sink>
   <expirationTime>2003-10-21T21:57:07.247Z</expirationTime>
  </subscribe>


<definitions name="OGSINotificationSubscription" 
targetNamespace="http://www.gridforum.org/namespaces/2003/03/OGSI/service" 
xmlns="http://schemas.xmlsoap.org/wsdl/" 
xmlns:ogsibinding="http://www.gridforum.org/namespaces/2003/03/OGSI/bindings" 
xmlns:soap="http://schemas.xmlsoap.org/wsdl/soap/">
  <import location="http://192.168.1.2:8081/schema/ogsi/ogsi_bindings.wsdl" 
    namespace="http://www.gridforum.org/namespaces/2003/03/OGSI/bindings"/>
  <service name="OGSINotificationSubscription" 
    xmlns:gsdl="http://ogsa.globus.org/">
    <gsdl:instanceOf handle="http://192.168.1.2:8081/ogsa/services/ogsi/NotificationSubscriptionFactoryService/hash-26675936-1066769829203"/>
    <gsdl:instanceOf handle="http://192.168.1.2:8081/ogsa/services/instance"/>
    <port binding="ogsibinding:NotificationSubscriptionSOAPBinding" name="NotificationSubscriptionPort">
        <soap:address location="http://192.168.1.2:8081/ogsa/services/ogsi/NotificationSubscriptionFactoryService/hash-26675936-1066769829203"/>
    </port>
  </service>
</definitions>    
    """
    subscription = None

    def __init__(self):
        GridServiceInterface.__init__(self)

    def subscribe(self, msg):
        post = '%s/hash-%s' %(post,time.time())
        sub = self.subscription(addr=self.addr, port=self.port, post=post, reference=None)
        sub.setSubscription(msg, self.serviceData)

        
class NotificationServiceGroupService(NotificationServiceGroupSOAPBinding, NotificationServiceInterface):
    """This is the ContainerRegistryService persistent service.
    <port binding="notificationservicegroupbinding:NotificationServiceGroupSOAPBinding" 
          name="NotificationServiceGroupPort">
        <soap:address location="http://localhost:8080/ogsa/services/"/>
    </port>
    """
    subscription = OGSINotificationSubscription

    def __init__(self, addr, port, post, reference):
        NotificationServiceGroupSOAPBinding.__init__(self, addr, port, post, reference)
        NotificationServiceInterface.__init__(self)
        self.setTerminationTime()
        self.setWSDLReference()
        """ 
        for sd in self.serviceData[ServiceGroup_portType.name.name].getInstances():
            sd.setSerializationElement(variable='serviceGroupEntryLocator', nil=True)
            sd.setSerializationElement(variable='memberServiceLocator')
            sd.setSerializationElement(variable='content')

            sd._memberServiceLocator._handle = 'http://crap'
            sd._content.setSerializationElement(variable='propertiesDetail',typecode=ns4.PropertiesDetailType_Def)
            sd._content._propertiesDetail.setSerializationElement(variable='factory')
            sd._content._propertiesDetail._name = 'not available'
            sd._content._propertiesDetail._portType = Holder()
            sd._content._propertiesDetail._state = "INACTIVE"
            sd._content._propertiesDetail._propertiesLifecycle = "TRANSIENT"
            sd._content._propertiesDetail._deploymentLifecycle = "TRANSIENT"
            sd._content._propertiesDetail._activateOnStartup = 0
            sd._content._propertiesDetail._handle = 'http://crap2'
        """ 

    def setEntry(self):
        logger.info('NotificationServiceGroupService.setEntry --\n\t%s', str(self.getDict()))

        for k,v in self.getDict().items():
            #if not isinstance(v, NotificationSource_portType):
            #    raise RuntimeError, 'all services must be instances of NotificationSource)'
            if k.endswith('HandleResolver'):
                continue
            logger.info('\t -- %s', k)
            self.serviceData[ServiceGroup_portType.name.name].addInstance()
            sd = self.serviceData[ServiceGroup_portType.name.name].getInstances()[-1]
            sd.setSerializationElement(variable='serviceGroupEntryLocator', nil=True)
            sd.setSerializationElement(variable='memberServiceLocator')
            sd.setSerializationElement(variable='content')
            sd._memberServiceLocator._handle = 'http://crap'
            sd._content.setSerializationElement(variable='propertiesDetail',typecode=ns4.PropertiesDetailType_Def)
            sd._content._propertiesDetail.setSerializationElement(variable='factory')
            sd._content._propertiesDetail._name = 'not available'
            sd._content._propertiesDetail._portType = Holder()
            sd._content._propertiesDetail._state = "INACTIVE"
            sd._content._propertiesDetail._propertiesLifecycle = "TRANSIENT"
            sd._content._propertiesDetail._deploymentLifecycle = "TRANSIENT"
            sd._content._propertiesDetail._activateOnStartup = 0
            sd._content._propertiesDetail._handle = 'http://crap2'
         
    def subscribe(self, msg):
        logger.info('NotificationServiceGroupService.subscribe')
        pyobj = NotificationServiceGroupSOAPBinding.subscribe(self, msg)
        pyobj.setSerializationElement(variable='subscriptionInstanceLocator')
        pyobj.setSerializationElement(variable='currentTerminationTime')
        pyobj._currentTerminationTime.setAttribute_before(time.gmtime(time.time()+10000))
        pyobj._subscriptionInstanceLocator.setSerializationElement(\
            variable='reference',typecode=ns1.WSDLReferenceType_Def)
        subscription = NotificationServiceInterface.subscribe(self, msg)
        pyobj._subscriptionInstanceLocator._reference._dom = subscription.getReference()
        return pyobj

    def findServiceData(self, msg):
        logger.info('NotificationServiceGroupService.findServiceData')

        pyobj = NotificationServiceGroupSOAPBinding.findServiceData(self, msg)
        pyobj.setSerializationElement(variable='result',)
        pyobj._result.setSerializationElement(variable='serviceDataValues',
            typecode=ns3.ServiceDataValuesElement_Dec)

        if msg and hasattr(msg,'_queryExpression'):
            names = msg._queryExpression._queryByServiceDataNames._name

            for name in names:
                if name == 'subscribeExtensibility':
                    pyobj._result._serviceDataValues._subscribeExtensibility  = None
                    break
                elif name == 'entry':
                    sd = self.serviceData[name].getInstances()[0]
                    logger.info('serviceData(%s) = %s', name, str(sd))
                    pyobj._result._serviceDataValues._entry = sd
                    #pyobj._result._serviceDataValues.setSerializationElement(variable=name,typecode=sd)

        return pyobj

