############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
#
###########################################################################

import ZSI, time, urlparse

from pyGridWare.ogsi.services.properties_detail_xsd_services_types import \
    ogsa_globus_org_types_properties as ns4

from pyGridWare.ogsi.services.OGSI_services import \
    ns1

from pyGridWare.ogsi.services.NotificationFactoryService_services import \
    FindServiceDataOutputMessageWrapper,\
    SubscribeOutputMessageWrapper,\
    CreateServiceOutputMessageWrapper

from pyGridWare.ogsi.services.NotificationFactoryService_services import \
    ns1 as ns2

from pyGridWare.ogsi.services.GWSDL_types import \
    www_gridforum_org_namespaces_2003_03_serviceData as ns3

from xml.dom.ext.reader import PyExpat
from xml.ns import WSDL
from Ft.Xml import XPath

# THIS IS GOING AWAY
from pyGridWare.ogsi.HandleResolvers import \
    SubscriptionReferenceDomAid


""" OGSIBaseServices

    This module defines some Base classes for services.

    GridService

    FactoryBase

"""

DEBUG = 0
def print_debug(msg, arg=None):
    if DEBUG and arg:
        print "<OGSIBaseServices>:  %s" %msg, arg
    elif DEBUG:
        print "<OGSIBaseServices>:  %s" %msg

class Holder: pass


class GridService:
    soapaction_dict = {}

    def __init__(self, addr, port, post, reference):
        #self.entries = []
        self.entries = None
        self.addr = addr
        self.port = port
        self.post = post
        self.reference = reference

    def addEntry(self, handle):
        """Not a callback function, for notifying CRS of new services.
        """
        key = '/ogsa/services/'
        properties_name = handle[(handle.find(key)+len(key)):]

        entry = ns2.EntryType_Def(name='entry')
        self.entries = entry

        entry._serviceGroupEntryLocator = ZSI.TC.Void('serviceGroupEntryLocator')
        entry._memberServiceLocator = ns2.LocatorType_Def(name="memberServiceLocator")
        entry._memberServiceLocator._handle = handle
        entry._content = ns2.EntryContentType_Def(name="content")
        entry._content._propertiesDetail = ns4.PropertiesDetailType_Def(name='propertiesDetail',typed=1)
        entry._content._propertiesDetail._name = properties_name
        entry._content._propertiesDetail._portType = Holder()
        entry._content._propertiesDetail._state = "INACTIVE"
        entry._content._propertiesDetail._propertiesLifecycle = "TRANSIENT"
        entry._content._propertiesDetail._deploymentLifecycle = "TRANSIENT"
        entry._content._propertiesDetail._activateOnStartup = 0
        entry._content._propertiesDetail._handle = handle

    def getReference(self):
        """Not a callback function, service returns a reference (service definition) of itself.
           <?xml version="1.0" encoding="UTF-8"?>
        """
        st = open(self.reference)
        reader = PyExpat.Reader()
        document = reader.fromStream(st)
        processorNss = {'wsdl':WSDL.BASE, 'soap':WSDL.BIND_SOAP}
        context = XPath.Context.Context(document, processorNss=processorNss)
        location = 'location'

        expr = 'wsdl:definitions//soap:*[@%s]' %location
        compExpr = XPath.Compile(expr)
        for e in compExpr.evaluate(context):
            scheme,netloc,path,params,query,fragment = \
                urlparse.urlparse(e.getAttribute(location))
            url_tuple = (scheme, 
                         '%s:%d' %(self.addr,self.port),
                         self.post,
                         params,
                         query,
                         fragment)
            url = urlparse.urlunparse(url_tuple)
            e.setAttribute(location, url)

        expr = 'wsdl:definitions/wsdl:import[@%s]' %location
        compExpr = XPath.Compile(expr)
        for e in compExpr.evaluate(context):
            scheme,netloc,path,params,query,fragment = \
                urlparse.urlparse(e.getAttribute(location))
            if not scheme:
                #support relative paths...
                scheme = 'http'
                endIdx = self.reference.rfind('/')
                if endIdx > 0:
                    abs_path = '%s/%s' %(self.reference[:endIdx],path)
                else: 
                    abs_path = path

            url_tuple = (scheme or 'http', 
                         '%s:%d' %(self.addr,self.port),
                         abs_path,
                         params,
                         query,
                         fragment)
            url = urlparse.urlunparse(url_tuple)
            e.setAttribute(location, url)

        return document

    def setServiceData(self, *args):
        #SetServiceDataInputMessage
        #SetServiceDataOutputMessageWrapper()
        print_debug("def setServiceData(*args):", args)
        return None

    def destroy(self, *args):
        #DestroyInputMessage
        #DestroyOutputMessageWrapper
        print_debug("def destroy(*args):", args)
        return None

    def findServiceData(self, msg):
        print_debug("GridService.findServiceData")
        response = FindServiceDataOutputMessageWrapper()
        response._result._serviceDataValues = ns3.ServiceDataValuesElement_Dec()

        #names = ['entry']
        if msg and hasattr(msg,'_queryExpression'):
            names = msg._queryExpression._queryByServiceDataNames._name

        for name in names:
            if name == 'subscribeExtensibility':
                response._result._serviceDataValues._subscribeExtensibility  = \
                    ns3.subscribeExtensibility_Def(name='subscribeExtensibility',
                                                   repeatable=1,
                                                   optional=1)
                break
            elif name == 'entry':
                response._result._serviceDataValues._entry = self.entries

        return response

    def requestTerminationBefore(self, *args):
        #RequestTerminationBeforeInputMessageWrapper()
        #RequestTerminationBeforeOutputMessageWrapper()
        print_debug("requestTerminationBefore(*args):", args)
        return None

    def requestTerminationAfter(self, *args):
        #RequestTerminationAfterInputMessage
        #RequestTerminationAfterOutputMessageWrapper()
        print_debug("def requestTerminationAfter(*args):", args)
        return None


class FactoryBase(GridService):
    """
       class variables:
          instance -- class definition of what we are creating.
    """
    instance = None
    instance_reference = None

    def __init__(self, addr, port, post, reference):
        GridService.__init__(self, addr, port, post, reference)
        self.soap_handler = None
        self.instances = {}

    def setSoapHandler(self, handler):
        self.soap_handler = handler

    def addInstance(self, instance):
        self.soap_handler.addService(instance)

    def createService(self, msg, handle=None, location=None):
        if not handle:
            handle = "http://%s:%d%s/hash-%s" \
                     %(self.addr,self.port,self.post,str(time.time()))
        if not location:
            location = handle

        hostport = urlparse.urlparse(handle)[1].split(':') + ['80']
        reply = CreateServiceOutputMessageWrapper()
        post = urlparse.urlparse(handle)[2]
        service = self.instance(
                      addr=self.addr, 
                      port=self.port, 
                      post=post, 
                      reference=self.instance_reference
                      )
        self.addInstance(service)

        self.addEntry(handle=handle)
        reply._currentTerminationTime = ns1.TerminationTimeType_Def(name="currentTerminationTime")
        reply._locator = ns2.LocatorType_Def(name="locator")
        reply._locator._reference = ns1.WSDLReferenceType_Def(name='reference')
        reply._extensibilityOutput = ZSI.TC.Void('extensibilityOutput')
        reply._locator._reference._dom = service.getReference()

        return reply


class NotificationSource(GridService):

    def __init__(self, addr, port, post, reference):
        GridService.__init__(self, addr, port, post, reference)

    def subscribe(self, msg):
        #SubscribeInputMessage
        #SubscribeOutputMessageWrapper()
        print_debug("def subscribe(msg):", msg)
        sub_ref = SubscriptionReferenceDomAid()
        return sub_ref.getSubscribeOutputMessage(msg, self.addr, self.port)


class NotificationSink(GridService):

    def __init__(self, addr, port, post, reference):
        GridService.__init__(self, addr, port, post, reference)

    def deliverNotification(self, msg):
        """this is a one-way operation
        """
        print_debug("NotificationSink.deliverNotification", msg)
        print_debug("\tserviceDataValues -- ", msg.serviceDataValues)
        print_debug("\t\tterminationTime -- ", msg.serviceDataValues._terminationTime)
        for k,v in request.message.serviceDataValues._terminationTime.attributes_dict.items():
            print_debug("\t\t\t%s  -- %s" %(k,v))
        return None



