############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
#
###########################################################################
import urlparse
from ZSI import _seqtypes

from pyGridWare.ogsi.services.OGSI_services import \
    FindByHandleOutputMessageWrapper, SetServiceDataOutputMessageWrapper, DestroyOutputMessageWrapper,\
    RequestTerminationAfterOutputMessageWrapper, FindServiceDataOutputMessageWrapper, \
    RequestTerminationBeforeOutputMessageWrapper, ns1

from pyGridWare.ogsi.services.GWSDL_types import \
    www_gridforum_org_namespaces_2003_03_serviceData as ns3

from pyGridWare.ogsi.TCmodified import AnyElement, AnyURI, Base64Binary, QName
from pyGridWare.ogsi.services.GridService import GridServiceInterface, GridService_portType, PortType


from pyGridWare import getLogger
logger = getLogger(__name__)
    
""" OGSIHandleResolverService

     This is one of the modules passed to dispatch for the stand-alone server.
    This module should define GridService interactions.

    HandleResolver_portType(GridService_portType)

    HandleResolverSOAPBinding  

    OGSIHandleResolverService

"""
class HandleResolver_gPortType(GridService_gPortType, PortType):
    """
    <gwsdl:portType name="HandleResolver" extends="ogsi:GridService">
        <operation name="findByHandle">
    """
    class handleResolverScheme(PortType.ServiceData):
        name="handleResolverScheme"
        type=AnyURI
        minOccurs=1
        maxOccurs="unbounded"
        mutability="mutable"
        modifiable=False
        nillable=True

    serviceDataTypes = GridService_portType.serviceDataTypes.copy()
    serviceDataTypes[handleResolverScheme.name] = handleResolverScheme

    def findByHandle(self, msg):
        raise NotImplementedError, 'portType is a interface'


class HandleResolverSOAPBinding(HandleResolver_gPortType):
    """
    <binding name="HandleResolverSOAPBinding" type="ogsi:HandleResolver">

    <port binding="ogsibinding:HandleResolverSOAPBinding" name="HandleResolverPort">
        <soap:address location="http://localhost:8080/ogsa/services/"/>
    </port>
    """
    soapAction = {\
        'http://www.gridforum.org/namespaces/2003/03/OGSI#findByHandle':'findByHandle',
        'http://www.gridforum.org/namespaces/2003/03/OGSI#setServiceData':'setServiceData',
        'http://www.gridforum.org/namespaces/2003/03/OGSI#destroy':'destroy',
        'http://www.gridforum.org/namespaces/2003/03/OGSI#requestTerminationAfter':'requestTerminationAfter',
        'http://www.gridforum.org/namespaces/2003/03/OGSI#findServiceData':'findServiceData',
        'http://www.gridforum.org/namespaces/2003/03/OGSI#requestTerminationBefore':'requestTerminationBefore',
        }

    def findByHandle(self, msg):
        logger.info('HandleResolverSOAPBinding.findByHandle')
        pyobj = FindByHandleOutputMessageWrapper()
        pyobj.setSerializationElement(variable='locator',)
        return pyobj

    def setServiceData(self, msg):
        logger.info('HandleResolverSOAPBinding.setServiceData')
        pyobj = SetServiceDataOutputMessageWrapper()
        return pyobj

    def destroy(self, msg):
        logger.info('HandleResolverSOAPBinding.destroy')
        pyobj = DestroyOutputMessageWrapper()
        return pyobj

    def requestTerminationAfter(self, msg):
        logger.info('HandleResolverSOAPBinding.requestTerminationAfter')
        pyobj = RequestTerminationAfterOutputMessageWrapper()
        return pyobj

    def findServiceData(self, msg):
        logger.info('HandleResolverSOAPBinding.findServiceData')
        pyobj = FindServiceDataOutputMessageWrapper()
        return pyobj

    def requestTerminationBefore(self, msg):
        logger.info('HandleResolverSOAPBinding.requestTerminationBefore')
        pyobj = RequestTerminationBeforeOutputMessageWrapper()
        return pyobj

class HandleResolverPort(HandleResolverSOAPBinding):
    pass


class OGSIHandleResolverService(Service):
    """A Serivce's callback methods should'nt be called directly,
       instead use the service instance as a function and pass
       the binding (soapAction) and message information as params.

          example: 
              service = OGSIHandleResolverService(..)
              pyobj = service(soapAction,msg)
    """
    def __init__(self, addr, port, post, reference):
        Service.__init__(self)
        self.ports['HandleResolverPort'] = HandleResolverPort(addr,port,post,reference)
        self.setTerminationTime()
        self.setWSDLReference()

    def findByHandle(self, port, bind, msg):
        """port -- port call goes out on.
           bind -- soapAction value, specifies operation.
           msg  -- pyobj, input message/typecode.
        """
        pyobj = port(bind, msg)
        logger.info('OGSIHandleResolverService.findByHandle')
        if msg:
            urls = msg._handleSet._handle
        else:
            urls = (self.post,)
        if not type(urls) in _seqtypes:
            handle = urls
        else:
            handle = urls[0]

        scheme,netloc,path,params,query,fragment = urlparse.urlparse(handle)
        try:
            service = self.getNode(path)
        except KeyError, ex:
            logger.error("missing GSH(%s)" %path)
            pyobj = None
        else:
            logger.info("found service GSH(%s)" %path)
            pyobj._locator._reference = service.getWSDLReference()
            pyobj._locator._handle = handle
        return pyobj

    def setServiceData(self, port, bind, msg):
        """port -- port call goes out on.
           bind -- soapAction value, specifies operation.
           msg  -- pyobj, input message/typecode.
        """
        pyobj = port(bind, msg)
        logger.info('OGSIHandleResolverService.setServiceData')
        self.setServiceData_SDE(msg._updateExpression)
        return pyobj

    def destroy(self, port, bind, msg):
        """port -- port call goes out on.
           bind -- soapAction value, specifies operation.
           msg  -- pyobj, input message/typecode.
        """
        pyobj = port(bind, msg)
        logger.info('OGSIHandleResolverService.destroy')
        self.destroy_SDE()
        return pyobj

    def requestTerminationAfter(self, port, bind, msg):
        """port -- port call goes out on.
           bind -- soapAction value, specifies operation.
           msg  -- pyobj, input message/typecode.
        """
        pyobj = port(bind, msg)
        logger.info('OGSIHandleResolverService.requestTerminationAfter')
        self.requestTerminationAfter_SDE(msg._terminationTime)
        pyobj._currentTerminationTime = self.requestTerminationAfter_SDE(msg._terminationTime)
        return pyobj

    def findServiceData(self, port, bind, msg):
        """port -- port call goes out on.
           bind -- soapAction value, specifies operation.
           msg  -- pyobj, input message/typecode.

           SDE:
             NotificationSource portType 
               --subscribeExtensibility
             ServiceGroup portType
               --entry 
               --membershipContentRule
        """
        logger.info('OGSIHandleResolverService.findServiceData')
        pyobj = port(bind, msg)
        pyobj.setSerializationElement(variable='result',)
        pyobj._result.setSerializationElement(variable='serviceDataValues', 
            typecode=ns3.ServiceDataValuesElement_Dec)

        #Query has an ExtensibilityType
        sd = GridService_gPortType.findServiceData(port, msg._queryExpression)
 
        #Check Static ServiceData 
        if hasattr(query, '_queryByServiceDataNames'):
            self.queryByServiceDataNames(pyobj=query._queryByServiceDataNames)
            #for name in query._queryByServiceDataNames._name:
            #    logger.info('query name = %s', name)
            #    var = '_%s' %name
            #    setattr(pyobj._result._serviceDataValues, var, getattr(port, name)())

        elif hasattr(query, '_setByServiceDataNames'):
            pass
        elif hasattr(query, '_deleteByServiceDataNames'):
            pass
        else:
             logger.error('findServiceData unknown inputElement in query')
             raise

        return pyobj

    def requestTerminationBefore(self, port, bind, msg):
        """port -- port call goes out on.
           bind -- soapAction value, specifies operation.
           msg  -- pyobj, input message/typecode.
        """
        pyobj = HandleResolverSOAPBinding.requestTerminationBefore(self, msg)
        logger.info('OGSIHandleResolverService.findServiceData.requestTerminationBefore')
        self.setTerminationTime(before=msg._terminationTime)
        pyobj._currentTerminationTime = self.getTerminationTime()
        return pyobj

