############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
#
###########################################################################
import urlparse, time
from ZSI import _seqtypes

from pyGridWare.ogsi.services.OGSI_services import \
    DeliverNotificationInputMessageWrapper, DestroyOutputMessageWrapper,\
    FindServiceDataOutputMessageWrapper, RequestTerminationAfterOutputMessageWrapper,\
    RequestTerminationBeforeOutputMessageWrapper, SetServiceDataOutputMessageWrapper,\
    SubscribeOutputMessageWrapper, OGSIServiceLocator, ns1

from pyGridWare.ogsi.services.GWSDL_types import \
    www_gridforum_org_namespaces_2003_03_serviceData as ns3

from properties_detail_xsd_services_types import ns1 as ns4

from pyGridWare.ogsi.TCmodified import AnyElement, AnyURI, Base64Binary, QName
from pyGridWare.ogsi.services.GridService import \
    NotificationServiceInterface, GridServiceInterface, PortType, GridService_portType

from pyGridWare import getLogger
logger = getLogger(__name__)
    
""" OGSINotificationSubscription

     This is one of the modules passed to dispatch for the stand-alone server.
    This module should define GridService interactions.

    classes: 
        NotificationSubscription_portType

        NotificationSubscriptionSOAPBinding

        OGSINotificationSubscription


"""

class NotificationSubscription_portType(GridService_portType):
    """<gwsdl:portType name="NotificationSubscription" extends="ogsi:GridService">
       This needs a deliverNotification operation, but it is not specified in the wsdl
    """
    class subscriptionExpression(PortType.ServiceData):
        name="subscriptionExpression"
        type=AnyElement
        minOccurs=1
        maxOccurs=1
        mutability="mutable"
        modifiable=False
        nillable=False
    class sinkLocator(PortType.ServiceData):
        name="sinkLocator"
        type=ns1.LocatorType_Def
        minOccurs=1
        maxOccurs=1
        mutability="mutable"
        modifiable=False
        nillable=False

    serviceDataTypes = GridService_portType.serviceDataTypes.copy()
    serviceDataTypes[subscriptionExpression.name] = subscriptionExpression
    serviceDataTypes[sinkLocator.name] = sinkLocator


class NotificationSubscriptionSOAPBinding(NotificationSubscription_portType):
    """<binding name="NotificationSubscriptionSOAPBinding" type="ogsi:NotificationSubscription">
    """
    soapAction = {\
        'http://www.gridforum.org/namespaces/2003/03/OGSI#setServiceData':'setServiceData',
        'http://www.gridforum.org/namespaces/2003/03/OGSI#destroy':'destroy',
        'http://www.gridforum.org/namespaces/2003/03/OGSI#requestTerminationAfter':'requestTerminationAfter',
        'http://www.gridforum.org/namespaces/2003/03/OGSI#findServiceData':'findServiceData',
        'http://www.gridforum.org/namespaces/2003/03/OGSI#requestTerminationBefore':'requestTerminationBefore',
        }
    def setServiceData(self, msg):
        logger.info('NotificationSubscriptionSOAPBinding.setServiceData')
        pyobj = SetServiceDataOutputMessageWrapper()
        return pyobj

    def destroy(self, msg):
        logger.info('NotificationSubscriptionSOAPBinding.destroy')
        pyobj = DestroyOutputMessageWrapper()
        return pyobj

    def requestTerminationAfter(self, msg):
        logger.info('NotificationSubscriptionSOAPBinding.requestTerminationAfter')
        pyobj = RequestTerminationAfterOutputMessageWrapper()
        return pyobj

    def findServiceData(self, msg):
        logger.info('NotificationSubscriptionSOAPBinding.findServiceData')
        pyobj = FindServiceDataOutputMessageWrapper()
        return pyobj

    def requestTerminationBefore(self, msg):
        logger.info('NotificationSubscriptionSOAPBinding.requestTerminationBefore')
        pyobj = RequestTerminationBeforeOutputMessageWrapper()
        return pyobj


class OGSINotificationSubscription(NotificationSubscriptionSOAPBinding, GridServiceInterface):
    """  ogsi_notification_subscription_service.wsdl

  <deliverNotification xmlns="http://www.gridforum.org/namespaces/2003/03/OGSI">
   <message>
    <ns1:serviceDataValues xmlns:ns1="http://www.gridforum.org/namespaces/2003/03/serviceData">
     <terminationTime xsi:type="ns2:TerminationTimeType" ns2:after="2003-08-13T05:25:54.732Z" ns2:before="2003-08-13T05:25:54.732Z" ns2:timestamp="2003-08-13T05:15:43.296Z" xmlns:ns2="http://www.gridforum.org/namespaces/2003/03/OGSI"/>
    </ns1:serviceDataValues>
   </message>
  </deliverNotification>
    """
    reference = None

    def __init__(self, addr, port, post, reference):
        """a single reference to a serviceData declaration
        """
        NotificationServiceGroupSOAPBinding.__init__(self, addr, port, post, reference)
        GridServiceInterface.__init__(self)
        self._serviceData = None
        self._modificationTime = None
        self._client_binding = None
        self._subscriber = None

    def setSubscription(self, msg, sd):
        scheme,netloc,path,params,query,fragment = urlparse.urlparse(msg._sink._handle)
        locator = OGSIServiceLocator()
        self._client_binding = locator.getNotificationSink(\
            portAddress=msg._sink._handle, **{'url':path})
        self.setNode(self,path)

        msg._expirationTime

        msg._subscriptionExpression
        msg._subscriptionExpression._subscribeByServiceDataNames
        name = msg._subscriptionExpression._subscribeByServiceDataNames._name
        self._serviceData = sd[name]
        self._modificationTime = self._serviceData.getModificationTime()

    def deliverNotification(self):
        msg = DeliverNotificationInputMessageWrapper()
        msg.setSerializationElement(variable='message')
        msg._message.setSerializationElement(\
            variable='serviceDataValues',
            typecode=ns3.ServiceDataValuesElement_Dec)
        setattr(msg._message._serviceDataValues,
            '_%s' %self._serviceData.name,
            self._serviceData.getInstances())
        self._client_binding.deliverNotification(msg)

    def check(self):
        #do a check to see if service data has changed
        if self._serviceData.getModificationTime() > self._modificationTime:
            logger.info('serviceData(%s) has changed', self._serviceData.name)
            self.deliverNotification()
        elif self._serviceData.getModificationTime < self._modificationTime:
            logger.error('serviceData(%s) bad modification time', self._serviceData.name)
        else:    
            logger.info('serviceData(%s) no change', self._serviceData.name)

